from dataclasses import dataclass

from .enums import ParseableEnum, ResourceType


class Priv(ParseableEnum):
    pass


@dataclass
class GrantedPrivilege:
    privilege: Priv
    on: str  # probably should be FQN

    @classmethod
    def from_grant(cls, privilege: str, granted_on: str, name: str):
        resource_type = ResourceType(granted_on)
        priv_type = PRIVS_FOR_RESOURCE_TYPE[resource_type]
        if priv_type:
            priv = priv_type(privilege)
            return cls(privilege=priv, on=name)
        else:
            return cls(privilege=privilege, on=granted_on)


class AccountPriv(Priv):
    ALL = "ALL"
    APPLY_AGGREGATION_POLICY = "APPLY AGGREGATION POLICY"
    APPLY_AUTHENTICATION_POLICY = "APPLY AUTHENTICATION POLICY"
    APPLY_MASKING_POLICY = "APPLY MASKING POLICY"
    APPLY_PACKAGES_POLICY = "APPLY PACKAGES POLICY"
    APPLY_PASSWORD_POLICY = "APPLY PASSWORD POLICY"
    APPLY_PROJECTION_POLICY = "APPLY PROJECTION POLICY"
    APPLY_RESOURCE_GROUP = "APPLY RESOURCE GROUP"
    APPLY_ROW_ACCESS_POLICY = "APPLY ROW ACCESS POLICY"
    APPLY_SESSION_POLICY = "APPLY SESSION POLICY"
    APPLY_TAG = "APPLY TAG"
    ATTACH_POLICY = "ATTACH POLICY"
    AUDIT = "AUDIT"
    BIND_SERVICE_ENDPOINT = "BIND SERVICE ENDPOINT"
    CANCEL_QUERY = "CANCEL QUERY"
    CREATE_ACCOUNT = "CREATE ACCOUNT"
    CREATE_API_INTEGRATION = "CREATE API INTEGRATION"
    CREATE_APPLICATION = "CREATE APPLICATION"
    CREATE_APPLICATION_PACKAGE = "CREATE APPLICATION PACKAGE"
    CREATE_COMPUTE_POOL = "CREATE COMPUTE POOL"
    CREATE_CREDENTIAL = "CREATE CREDENTIAL"
    CREATE_DATA_EXCHANGE_LISTING = "CREATE DATA EXCHANGE LISTING"
    CREATE_DATABASE = "CREATE DATABASE"
    CREATE_EXTERNAL_VOLUME = "CREATE EXTERNAL VOLUME"
    CREATE_FAILOVER_GROUP = "CREATE FAILOVER GROUP"
    CREATE_INTEGRATION = "CREATE INTEGRATION"
    CREATE_NETWORK_POLICY = "CREATE NETWORK POLICY"
    CREATE_REPLICATION_GROUP = "CREATE REPLICATION GROUP"
    CREATE_ROLE = "CREATE ROLE"
    CREATE_SHARE = "CREATE SHARE"
    CREATE_USER = "CREATE USER"
    CREATE_WAREHOUSE = "CREATE WAREHOUSE"
    EXECUTE_ALERT = "EXECUTE ALERT"
    EXECUTE_DATA_METRIC_FUNCTION = "EXECUTE DATA METRIC FUNCTION"
    EXECUTE_MANAGED_ALERT = "EXECUTE MANAGED ALERT"
    EXECUTE_MANAGED_TASK = "EXECUTE MANAGED TASK"
    EXECUTE_TASK = "EXECUTE TASK"
    IMPORT_SHARE = "IMPORT SHARE"
    MANAGE_ACCOUNT_SUPPORT_CASES = "MANAGE ACCOUNT SUPPORT CASES"
    MANAGE_EVENT_SHARING = "MANAGE EVENT SHARING"
    MANAGE_GRANTS = "MANAGE GRANTS"
    MANAGE_USER_SUPPORT_CASES = "MANAGE USER SUPPORT CASES"
    MANAGE_WAREHOUSES = "MANAGE WAREHOUSES"
    MODIFY_LOG_LEVEL = "MODIFY LOG LEVEL"
    MODIFY_METRIC_LEVEL = "MODIFY METRIC LEVEL"
    MODIFY_SESSION_LOG_LEVEL = "MODIFY SESSION LOG LEVEL"
    MODIFY_SESSION_METRIC_LEVEL = "MODIFY SESSION METRIC LEVEL"
    MODIFY_SESSION_TRACE_LEVEL = "MODIFY SESSION TRACE LEVEL"
    MODIFY_TRACE_LEVEL = "MODIFY TRACE LEVEL"
    MONITOR = "MONITOR"
    MONITOR_EXECUTION = "MONITOR EXECUTION"
    MONITOR_SECURITY = "MONITOR SECURITY"
    MONITOR_USAGE = "MONITOR USAGE"
    OVERRIDE_SHARE_RESTRICTIONS = "OVERRIDE SHARE RESTRICTIONS"
    PURCHASE_DATA_EXCHANGE_LISTING = "PURCHASE DATA EXCHANGE LISTING"
    RESOLVE_ALL = "RESOLVE ALL"


class AlertPriv(ParseableEnum):
    ALL = "ALL"
    MONITOR = "MONITOR"
    OPERATE = "OPERATE"
    OWNERSHIP = "OWNERSHIP"


class DatabasePriv(ParseableEnum):
    ALL = "ALL"
    APPLYBUDGET = "APPLYBUDGET"
    CREATE_DATABASE_ROLE = "CREATE DATABASE ROLE"
    CREATE_SCHEMA = "CREATE SCHEMA"
    # IMPORTED_PRIVILEGES = "IMPORTED PRIVILEGES" # only granted on shared database
    MODIFY = "MODIFY"
    MONITOR = "MONITOR"
    OWNERSHIP = "OWNERSHIP"
    REFERENCE_USAGE = "REFERENCE_USAGE"  # Cannot be granted to roles, only shares
    USAGE = "USAGE"


class DatabaseRolePriv(ParseableEnum):
    OWNERSHIP = "OWNERSHIP"
    USAGE = "USAGE"


class DirectoryTablePriv(ParseableEnum):
    OWNERSHIP = "OWNERSHIP"


class EventTablePriv(ParseableEnum):
    ALL = "ALL"
    OWNERSHIP = "OWNERSHIP"
    SELECT = "SELECT"
    INSERT = "INSERT"


class FailoverGroupPriv(ParseableEnum):
    ALL = "ALL"
    FAILOVER = "FAILOVER"
    MODIFY = "MODIFY"
    MONITOR = "MONITOR"
    OWNERSHIP = "OWNERSHIP"
    REPLICATE = "REPLICATE"


class FileFormatPriv(ParseableEnum):
    ALL = "ALL"
    OWNERSHIP = "OWNERSHIP"
    USAGE = "USAGE"


class FunctionPriv(ParseableEnum):
    ALL = "ALL"
    OWNERSHIP = "OWNERSHIP"
    USAGE = "USAGE"


class IntegrationPriv(ParseableEnum):
    ALL = "ALL"
    USAGE = "USAGE"
    USE_ANY_ROLE = "USE_ANY_ROLE"
    OWNERSHIP = "OWNERSHIP"


class MaterializedViewPriv(ParseableEnum):
    ALL = "ALL"
    APPLYBUDGET = "APPLYBUDGET"
    OWNERSHIP = "OWNERSHIP"
    REFERENCES = "REFERENCES"
    SELECT = "SELECT"


class NetworkPolicyPriv(ParseableEnum):
    OWNERSHIP = "OWNERSHIP"


class NetworkRulePriv(ParseableEnum):
    OWNERSHIP = "OWNERSHIP"


class NotebookPriv(ParseableEnum):
    OWNERSHIP = "OWNERSHIP"


class PackagesPolicyPriv(ParseableEnum):
    OWNERSHIP = "OWNERSHIP"
    USAGE = "USAGE"


class PasswordPolicyPriv(ParseableEnum):
    OWNERSHIP = "OWNERSHIP"


class PipePriv(ParseableEnum):
    ALL = "ALL"
    APPLYBUDGET = "APPLYBUDGET"
    MONITOR = "MONITOR"
    OPERATE = "OPERATE"
    OWNERSHIP = "OWNERSHIP"


class ProcedurePriv(ParseableEnum):
    ALL = "ALL"
    OWNERSHIP = "OWNERSHIP"
    USAGE = "USAGE"


class ReplicationGroupPriv(ParseableEnum):
    ALL = "ALL"
    MODIFY = "MODIFY"
    MONITOR = "MONITOR"
    OWNERSHIP = "OWNERSHIP"
    REPLICATE = "REPLICATE"


class RolePriv(ParseableEnum):
    OWNERSHIP = "OWNERSHIP"
    USAGE = "USAGE"


class SchemaPriv(ParseableEnum):
    ALL = "ALL"
    ADD_SEARCH_OPTIMIZATION = "ADD SEARCH OPTIMIZATION"
    APPLYBUDGET = "APPLYBUDGET"
    CREATE_AGGREGATION_POLICY = "CREATE AGGREGATION POLICY"
    CREATE_ALERT = "CREATE ALERT"
    CREATE_AUTHENTICATION_POLICY = "CREATE AUTHENTICATION POLICY"
    CREATE_CORTEX_SEARCH_SERVICE = "CREATE CORTEX SEARCH SERVICE"
    CREATE_DATASET = "CREATE DATASET"
    CREATE_DYNAMIC_TABLE = "CREATE DYNAMIC TABLE"
    CREATE_EVENT_TABLE = "CREATE EVENT TABLE"
    CREATE_EXTERNAL_TABLE = "CREATE EXTERNAL TABLE"
    CREATE_FILE_FORMAT = "CREATE FILE FORMAT"
    CREATE_FUNCTION = "CREATE FUNCTION"
    CREATE_GIT_REPOSITORY = "CREATE GIT REPOSITORY"
    CREATE_ICEBERG_TABLE = "CREATE ICEBERG TABLE"
    CREATE_IMAGE_REPOSITORY = "CREATE IMAGE REPOSITORY"
    CREATE_MASKING_POLICY = "CREATE MASKING POLICY"
    CREATE_MATERIALIZED_VIEW = "CREATE MATERIALIZED VIEW"
    CREATE_MODEL = "CREATE MODEL"
    CREATE_NETWORK_RULE = "CREATE NETWORK RULE"
    CREATE_NOTEBOOK = "CREATE NOTEBOOK"
    CREATE_PACKAGES_POLICY = "CREATE PACKAGES POLICY"
    CREATE_PASSWORD_POLICY = "CREATE PASSWORD POLICY"
    CREATE_PIPE = "CREATE PIPE"
    CREATE_PROCEDURE = "CREATE PROCEDURE"
    CREATE_PROJECTION_POLICY = "CREATE PROJECTION POLICY"
    CREATE_RESOURCE_GROUP = "CREATE RESOURCE GROUP"
    CREATE_ROW_ACCESS_POLICY = "CREATE ROW ACCESS POLICY"
    CREATE_SECRET = "CREATE SECRET"
    CREATE_SEQUENCE = "CREATE SEQUENCE"
    CREATE_SERVICE = "CREATE SERVICE"
    CREATE_SERVICE_CLASS = "CREATE SERVICE CLASS"
    CREATE_SESSION_POLICY = "CREATE SESSION POLICY"
    CREATE_SNAPSHOT = "CREATE SNAPSHOT"
    CREATE_STAGE = "CREATE STAGE"
    CREATE_STREAM = "CREATE STREAM"
    CREATE_STREAMLIT = "CREATE STREAMLIT"
    CREATE_TABLE = "CREATE TABLE"
    CREATE_TAG = "CREATE TAG"
    CREATE_TASK = "CREATE TASK"
    CREATE_TEMPORARY_TABLE = "CREATE TEMPORARY TABLE"
    CREATE_VIEW = "CREATE VIEW"
    MODIFY = "MODIFY"
    MONITOR = "MONITOR"
    OWNERSHIP = "OWNERSHIP"
    USAGE = "USAGE"


class SecretPriv(ParseableEnum):
    OWNERSHIP = "OWNERSHIP"
    READ = "READ"
    USAGE = "USAGE"


class SequencePriv(ParseableEnum):
    ALL = "ALL"
    OWNERSHIP = "OWNERSHIP"
    USAGE = "USAGE"


class StagePriv(ParseableEnum):
    ALL = "ALL"
    OWNERSHIP = "OWNERSHIP"
    READ = "READ"
    USAGE = "USAGE"
    WRITE = "WRITE"


class StreamPriv(ParseableEnum):
    ALL = "ALL"
    OWNERSHIP = "OWNERSHIP"
    SELECT = "SELECT"


class TablePriv(ParseableEnum):
    ALL = "ALL"
    APPLYBUDGET = "APPLYBUDGET"
    DELETE = "DELETE"
    EVOLVE_SCHEMA = "EVOLVE SCHEMA"
    INSERT = "INSERT"
    OWNERSHIP = "OWNERSHIP"
    REBUILD = "REBUILD"
    REFERENCES = "REFERENCES"
    SELECT = "SELECT"
    TRUNCATE = "TRUNCATE"
    UPDATE = "UPDATE"


class TagPriv(ParseableEnum):
    APPLY = "APPLY"
    OWNERSHIP = "OWNERSHIP"
    READ = "READ"


class TaskPriv(ParseableEnum):
    ALL = "ALL"
    APPLYBUDGET = "APPLYBUDGET"
    MONITOR = "MONITOR"
    OPERATE = "OPERATE"
    OWNERSHIP = "OWNERSHIP"


class UserPriv(ParseableEnum):
    ALL = "ALL"
    MONITOR = "MONITOR"
    OWNERSHIP = "OWNERSHIP"


class ViewPriv(ParseableEnum):
    ALL = "ALL"
    APPLYBUDGET = "APPLYBUDGET"
    DELETE = "DELETE"
    EVOLVE_SCHEMA = "EVOLVE SCHEMA"
    INSERT = "INSERT"
    OWNERSHIP = "OWNERSHIP"
    REBUILD = "REBUILD"
    REFERENCES = "REFERENCES"
    SELECT = "SELECT"
    TRUNCATE = "TRUNCATE"
    UPDATE = "UPDATE"


class WarehousePriv(ParseableEnum):
    ALL = "ALL"
    APPLYBUDGET = "APPLYBUDGET"
    MODIFY = "MODIFY"
    MONITOR = "MONITOR"
    OPERATE = "OPERATE"
    OWNERSHIP = "OWNERSHIP"
    USAGE = "USAGE"


PRIVS_FOR_RESOURCE_TYPE = {
    ResourceType.ACCOUNT: AccountPriv,
    ResourceType.AGGREGATION_POLICY: None,
    ResourceType.ALERT: AlertPriv,
    ResourceType.API_INTEGRATION: IntegrationPriv,
    ResourceType.APPLICATION_ROLE: None,
    ResourceType.AUTHENTICATION_POLICY: None,
    ResourceType.CATALOG_INTEGRATION: None,
    ResourceType.CLASS: None,
    ResourceType.COLUMN: None,
    ResourceType.COMPUTE_POOL: None,
    ResourceType.DATABASE_ROLE: DatabaseRolePriv,
    ResourceType.DATABASE: DatabasePriv,
    ResourceType.DIRECTORY_TABLE: DirectoryTablePriv,
    ResourceType.DYNAMIC_TABLE: TablePriv,
    ResourceType.EVENT_TABLE: EventTablePriv,
    ResourceType.EXTERNAL_ACCESS_INTEGRATION: IntegrationPriv,
    ResourceType.EXTERNAL_FUNCTION: FunctionPriv,
    ResourceType.FAILOVER_GROUP: FailoverGroupPriv,
    ResourceType.FILE_FORMAT: FileFormatPriv,
    ResourceType.FUNCTION: FunctionPriv,
    ResourceType.FUTURE_GRANT: None,
    ResourceType.GIT_REPOSITORY: None,
    ResourceType.GRANT_ON_ALL: None,
    ResourceType.GRANT: None,
    ResourceType.HYBRID_TABLE: None,
    ResourceType.IMAGE_REPOSITORY: None,
    ResourceType.INTEGRATION: IntegrationPriv,
    ResourceType.MATERIALIZED_VIEW: MaterializedViewPriv,
    ResourceType.NETWORK_POLICY: NetworkPolicyPriv,
    ResourceType.NETWORK_RULE: NetworkRulePriv,
    ResourceType.NOTEBOOK: NotebookPriv,
    ResourceType.NOTIFICATION_INTEGRATION: IntegrationPriv,
    ResourceType.PACKAGES_POLICY: PackagesPolicyPriv,
    ResourceType.PASSWORD_POLICY: PasswordPolicyPriv,
    ResourceType.PIPE: PipePriv,
    ResourceType.PROCEDURE: ProcedurePriv,
    ResourceType.REPLICATION_GROUP: ReplicationGroupPriv,
    ResourceType.RESOURCE_MONITOR: None,
    ResourceType.ROLE_GRANT: None,
    ResourceType.ROLE: RolePriv,
    ResourceType.SCHEMA: SchemaPriv,
    ResourceType.SECRET: SecretPriv,
    ResourceType.SECURITY_INTEGRATION: IntegrationPriv,
    ResourceType.SEQUENCE: SequencePriv,
    ResourceType.SERVICE: None,
    ResourceType.SHARE: None,
    ResourceType.STAGE: StagePriv,
    ResourceType.STORAGE_INTEGRATION: IntegrationPriv,
    ResourceType.STREAM: StreamPriv,
    ResourceType.TABLE: TablePriv,
    ResourceType.TAG_REFERENCE: None,
    ResourceType.TAG: TagPriv,
    ResourceType.TASK: TaskPriv,
    ResourceType.USER: UserPriv,
    ResourceType.VIEW: ViewPriv,
    ResourceType.WAREHOUSE: WarehousePriv,
}


CREATE_PRIV_FOR_RESOURCE_TYPE: dict[ResourceType, ParseableEnum] = {
    ResourceType.ACCOUNT: AccountPriv.CREATE_ACCOUNT,
    ResourceType.ALERT: SchemaPriv.CREATE_ALERT,
    ResourceType.API_INTEGRATION: AccountPriv.CREATE_API_INTEGRATION,
    ResourceType.DATABASE: AccountPriv.CREATE_DATABASE,
    ResourceType.DYNAMIC_TABLE: SchemaPriv.CREATE_DYNAMIC_TABLE,
    ResourceType.EVENT_TABLE: SchemaPriv.CREATE_TABLE,
    ResourceType.EXTERNAL_ACCESS_INTEGRATION: AccountPriv.CREATE_INTEGRATION,
    ResourceType.EXTERNAL_FUNCTION: SchemaPriv.CREATE_FUNCTION,
    ResourceType.FAILOVER_GROUP: AccountPriv.CREATE_FAILOVER_GROUP,
    ResourceType.FILE_FORMAT: SchemaPriv.CREATE_FILE_FORMAT,
    ResourceType.FUNCTION: SchemaPriv.CREATE_FUNCTION,
    # ResourceType.GRANT: AccountPriv.CREATE_GRANT,
    ResourceType.MATERIALIZED_VIEW: SchemaPriv.CREATE_MATERIALIZED_VIEW,
    ResourceType.NETWORK_POLICY: AccountPriv.CREATE_NETWORK_POLICY,
    ResourceType.NETWORK_RULE: SchemaPriv.CREATE_NETWORK_RULE,
    ResourceType.PACKAGES_POLICY: SchemaPriv.CREATE_PACKAGES_POLICY,
    ResourceType.PASSWORD_POLICY: SchemaPriv.CREATE_PASSWORD_POLICY,
    ResourceType.PIPE: SchemaPriv.CREATE_PIPE,
    ResourceType.PROCEDURE: SchemaPriv.CREATE_PROCEDURE,
    # ResourceType.RESOURCE_MONITOR: AccountPriv.CREATE_RESOURCE_MONITOR, # only ACCOUNTADMIN
    ResourceType.REPLICATION_GROUP: AccountPriv.CREATE_REPLICATION_GROUP,
    ResourceType.ROLE: AccountPriv.CREATE_ROLE,
    # ResourceType.ROLE_GRANT: RolePriv.OWNERSHIP,
    ResourceType.SCHEMA: DatabasePriv.CREATE_SCHEMA,
    ResourceType.SECRET: SchemaPriv.CREATE_SECRET,
    ResourceType.SEQUENCE: SchemaPriv.CREATE_SEQUENCE,
    ResourceType.STAGE: SchemaPriv.CREATE_STAGE,
    ResourceType.STREAM: SchemaPriv.CREATE_STREAM,
    ResourceType.TABLE: SchemaPriv.CREATE_TABLE,
    ResourceType.TAG: SchemaPriv.CREATE_TAG,
    ResourceType.TASK: SchemaPriv.CREATE_TASK,
    ResourceType.USER: AccountPriv.CREATE_USER,
    ResourceType.VIEW: SchemaPriv.CREATE_VIEW,
    ResourceType.WAREHOUSE: AccountPriv.CREATE_WAREHOUSE,
}


GLOBAL_PRIV_DEFAULT_OWNERS = {
    AccountPriv.APPLY_AGGREGATION_POLICY: "ACCOUNTADMIN",
    AccountPriv.APPLY_AUTHENTICATION_POLICY: "SECURITYADMIN",
    AccountPriv.APPLY_MASKING_POLICY: "ACCOUNTADMIN",
    AccountPriv.APPLY_PACKAGES_POLICY: "SECURITYADMIN",
    AccountPriv.APPLY_PASSWORD_POLICY: "SECURITYADMIN",
    AccountPriv.APPLY_PROJECTION_POLICY: "ACCOUNTADMIN",
    AccountPriv.APPLY_RESOURCE_GROUP: "ACCOUNTADMIN",
    AccountPriv.APPLY_ROW_ACCESS_POLICY: "ACCOUNTADMIN",
    AccountPriv.APPLY_SESSION_POLICY: "SECURITYADMIN",
    AccountPriv.APPLY_TAG: "ACCOUNTADMIN",
    AccountPriv.ATTACH_POLICY: "SECURITYADMIN",
    AccountPriv.AUDIT: "ACCOUNTADMIN",
    AccountPriv.BIND_SERVICE_ENDPOINT: "SECURITYADMIN",
    AccountPriv.CANCEL_QUERY: "ACCOUNTADMIN",
    AccountPriv.CREATE_ACCOUNT: "ACCOUNTADMIN",
    AccountPriv.CREATE_API_INTEGRATION: "ACCOUNTADMIN",
    AccountPriv.CREATE_APPLICATION_PACKAGE: "ACCOUNTADMIN",
    AccountPriv.CREATE_APPLICATION: "ACCOUNTADMIN",
    AccountPriv.CREATE_COMPUTE_POOL: "SYSADMIN",
    AccountPriv.CREATE_CREDENTIAL: "ACCOUNTADMIN",
    AccountPriv.CREATE_DATA_EXCHANGE_LISTING: "ACCOUNTADMIN",
    AccountPriv.CREATE_DATABASE: "SYSADMIN",
    AccountPriv.CREATE_EXTERNAL_VOLUME: "ACCOUNTADMIN",
    AccountPriv.CREATE_FAILOVER_GROUP: "ACCOUNTADMIN",
    AccountPriv.CREATE_INTEGRATION: "ACCOUNTADMIN",
    AccountPriv.CREATE_NETWORK_POLICY: "SECURITYADMIN",
    AccountPriv.CREATE_REPLICATION_GROUP: "ACCOUNTADMIN",
    AccountPriv.CREATE_ROLE: "USERADMIN",
    AccountPriv.CREATE_SHARE: "ACCOUNTADMIN",
    AccountPriv.CREATE_USER: "USERADMIN",
    AccountPriv.CREATE_WAREHOUSE: "SYSADMIN",
    AccountPriv.EXECUTE_ALERT: "ACCOUNTADMIN",
    AccountPriv.EXECUTE_DATA_METRIC_FUNCTION: "ACCOUNTADMIN",
    AccountPriv.EXECUTE_MANAGED_ALERT: "ACCOUNTADMIN",
    AccountPriv.EXECUTE_MANAGED_TASK: "ACCOUNTADMIN",
    AccountPriv.EXECUTE_TASK: "ACCOUNTADMIN",
    AccountPriv.IMPORT_SHARE: "ACCOUNTADMIN",
    AccountPriv.MANAGE_ACCOUNT_SUPPORT_CASES: "ACCOUNTADMIN",
    AccountPriv.MANAGE_EVENT_SHARING: "ACCOUNTADMIN",
    AccountPriv.MANAGE_GRANTS: "SECURITYADMIN",
    AccountPriv.MANAGE_USER_SUPPORT_CASES: "ACCOUNTADMIN",
    AccountPriv.MANAGE_WAREHOUSES: "ACCOUNTADMIN",
    AccountPriv.MODIFY_LOG_LEVEL: "ACCOUNTADMIN",
    AccountPriv.MODIFY_METRIC_LEVEL: "ACCOUNTADMIN",
    AccountPriv.MODIFY_SESSION_LOG_LEVEL: "ACCOUNTADMIN",
    AccountPriv.MODIFY_SESSION_METRIC_LEVEL: "ACCOUNTADMIN",
    AccountPriv.MODIFY_SESSION_TRACE_LEVEL: "ACCOUNTADMIN",
    AccountPriv.MODIFY_TRACE_LEVEL: "ACCOUNTADMIN",
    AccountPriv.MONITOR_EXECUTION: "ACCOUNTADMIN",
    AccountPriv.MONITOR_SECURITY: "ACCOUNTADMIN",
    AccountPriv.MONITOR_USAGE: "ACCOUNTADMIN",
    AccountPriv.MONITOR: "ACCOUNTADMIN",
    AccountPriv.OVERRIDE_SHARE_RESTRICTIONS: "ACCOUNTADMIN",
    AccountPriv.PURCHASE_DATA_EXCHANGE_LISTING: "ACCOUNTADMIN",
    AccountPriv.RESOLVE_ALL: "ACCOUNTADMIN",
}


def is_ownership_priv(priv):
    return str(priv) == "OWNERSHIP"


def _all_privs_for_resource_type(resource_type):
    all_privs = []
    for priv in PRIVS_FOR_RESOURCE_TYPE[resource_type]:
        priv = str(priv)
        if priv != "ALL" and priv != "OWNERSHIP":
            all_privs.append(priv)
    return all_privs


def execution_role_for_priv(priv: str):
    try:
        priv = AccountPriv(priv)
    except ValueError:
        return None
    return GLOBAL_PRIV_DEFAULT_OWNERS.get(priv)



import json

import click
import yaml

from titan.blueprint import dump_plan
from titan.operations.blueprint import blueprint_apply, blueprint_plan
from titan.operations.export import export_resources

from .identifiers import resource_type_for_label


def load_config(config_file):
    with open(config_file, "r") as f:
        config = yaml.safe_load(f)
    return config


def load_plan(plan_file):
    with open(plan_file, "r") as f:
        plan = json.load(f)
    return plan


@click.group()
def titan_cli():
    """titan core helps you manage your Snowflake environment."""
    pass


@titan_cli.command("plan", no_args_is_help=True)
@click.option("--config", "config_file", type=str, help="Path to configuration YAML file", metavar="<filename>")
@click.option("--json", "json_output", is_flag=True, help="Output plan in machine-readable JSON format")
@click.option("--out", "output_file", type=str, help="Write plan to a file", metavar="<filename>")
@click.option(
    "--mode",
    "run_mode",
    type=click.Choice(["CREATE-OR-UPDATE"]),
    metavar="<run_mode>",
    default="CREATE-OR-UPDATE",
    show_default=True,
    help="Run strategy",
)
def plan(config_file, json_output, output_file, run_mode):
    """Generate an execution plan based on your configuration"""
    config = load_config(config_file)
    plan_obj = blueprint_plan(config, run_mode)
    output = None
    if json_output:
        output = dump_plan(plan_obj, format="json")
    else:
        output = dump_plan(plan_obj, format="text")
    if output_file:
        with open(output_file, "w") as f:
            f.write(output)
    else:
        print(output)


@titan_cli.command("apply", no_args_is_help=True)
@click.option("--config", "config_file", type=str, help="Path to configuration YAML file", metavar="<filename>")
@click.option("--plan", "plan_file", type=str, help="Path to plan JSON file", metavar="<filename>")
@click.option(
    "--mode",
    "run_mode",
    type=click.Choice(["CREATE-OR-UPDATE"]),
    default="CREATE-OR-UPDATE",
    metavar="<run_mode>",
    show_default=True,
    help="Run strategy",
)
@click.option("--dry-run", is_flag=True, help="Perform a dry run without applying changes")
def apply(config_file, plan_file, run_mode, dry_run):
    """Apply a plan to Titan resources"""
    if config_file and plan_file:
        raise click.UsageError("Cannot specify both --config and --plan.")
    if not config_file and not plan_file:
        raise click.UsageError("Either --config or --plan must be specified.")
    if config_file:
        config = load_config(config_file)
        plan_obj = blueprint_plan(config, run_mode)
        blueprint_apply(plan_obj, run_mode, dry_run)
    else:
        plan_obj = load_plan(plan_file)
        blueprint_apply(plan_obj, run_mode, dry_run)


def _parse_resources(resource_labels_str):
    if resource_labels_str is None:
        return None
    return [resource_type_for_label(resource_label) for resource_label in resource_labels_str.split(",")]


@titan_cli.command("export", context_settings={"show_default": True}, no_args_is_help=True)
@click.option("--resource", type=str, help="The resource types to export", metavar="<resource_types>")
@click.option("--all", "export_all", is_flag=True, help="Export all resources")
@click.option("--exclude", "exclude", type=str, help="Exclude resources, used with --all", metavar="<resource_types>")
@click.option("--out", type=str, help="Write exported config to a file", metavar="<filename>")
@click.option("--format", type=click.Choice(["json", "yml"]), default="yml", help="Output format")
def export(resource, export_all, exclude, out, format):
    """
    This command allows you to export resources from Titan in either JSON or YAML format.
    You can specify the type of resource to export and the output filename for the exported data.

    Resource types are specified with snake case (eg. Warehouse => warehouse, NetworkRule => network_rule, etc.).
    You can also export all resources by using the --all flag. Specify multiple resource types by separating them with commas.

    Examples:

    \b
    # Export all database configurations to a file
    titan export --resource=database --out=databases.yml --format=yml

    \b
    # Export all resources
    titan export --all --out=titan.yml --format=yml

    \b
    # Export all resources except for users and roles
    titan export --all --exclude=user,role --out=titan.yml
    """

    if resource and export_all:
        raise click.UsageError("You can't specify both --resource and --all options at the same time.")

    resource_config = {}
    if resource:
        resource_config = export_resources(include=_parse_resources(resource))
    elif export_all:
        resource_config = export_resources(exclude=_parse_resources(exclude))
    else:
        raise

    output = None
    if format == "json":
        output = json.dumps(resource_config, indent=2)
    elif format == "yml":
        output = yaml.dump(resource_config, sort_keys=False)
    else:
        raise ValueError(f"Unsupported format: {format}")

    if out:
        with open(out, "w") as f:
            f.write(output)
    else:
        print(output)


if __name__ == "__main__":
    titan_cli()



from dataclasses import dataclass, field

from .resource import Resource, ResourceSpec, NamedResource
from .role import Role
from ..enums import AccountEdition, ParseableEnum, ResourceType
from ..props import (
    IntProp,
    StringProp,
    BoolProp,
    Props,
    EnumProp,
)
from ..resource_name import ResourceName
from ..scope import AccountScope


class InstanceFamily(ParseableEnum):
    CPU_X64_XS = "CPU_X64_XS"
    CPU_X64_S = "CPU_X64_S"
    CPU_X64_M = "CPU_X64_M"
    CPU_X64_L = "CPU_X64_L"
    CPU_X64_XL = "CPU_X64_XL"
    CPU_X64_2XL = "CPU_X64_2XL"
    CPU_X64_3XL = "CPU_X64_3XL"
    CPU_X64_4XL = "CPU_X64_4XL"


@dataclass(unsafe_hash=True)
class _ComputePool(ResourceSpec):
    name: ResourceName
    owner: Role = "ACCOUNTADMIN"
    min_nodes: int = None
    max_nodes: int = None
    instance_family: InstanceFamily = None
    auto_resume: bool = True
    initially_suspended: bool = field(default_factory=None, metadata={"fetchable": False})
    auto_suspend_secs: int = 3600
    comment: str = None


class ComputePool(NamedResource, Resource):
    """
    Description:
        A compute pool is a group of compute resources in Snowflake that can be used to execute SQL queries.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-compute-pool

    Fields:
        name (string, required): The unique name of the compute pool.
        owner (string or Role): The owner of the compute pool. Defaults to "ACCOUNTADMIN".
        min_nodes (int): The minimum number of nodes in the compute pool.
        max_nodes (int): The maximum number of nodes in the compute pool.
        instance_family (string or InstanceFamily): The family of instances to use for the compute nodes.
        auto_resume (bool): Whether the compute pool should automatically resume when queries are submitted. Defaults to True.
        initially_suspended (bool): Whether the compute pool should start in a suspended state.
        auto_suspend_secs (int): The number of seconds of inactivity after which the compute pool should automatically suspend. Defaults to 3600.
        comment (string): An optional comment about the compute pool.

    Python:
        ```python
        compute_pool = ComputePool(
            name="some_compute_pool",
            owner="ACCOUNTADMIN",
            min_nodes=2,
            max_nodes=10,
            instance_family="CPU_X64_S",
            auto_resume=True,
            initially_suspended=False,
            auto_suspend_secs=1800,
            comment="Example compute pool"
        )
        ```

    Yaml:
        ```yaml
        compute_pools:
          - name: some_compute_pool
            owner: ACCOUNTADMIN
            min_nodes: 2
            max_nodes: 10
            instance_family: CPU_X64_S
            auto_resume: true
            initially_suspended: false
            auto_suspend_secs: 1800
            comment: Example compute pool
        ```
    """

    edition = {AccountEdition.ENTERPRISE, AccountEdition.BUSINESS_CRITICAL}
    resource_type = ResourceType.COMPUTE_POOL
    props = Props(
        min_nodes=IntProp("min_nodes"),
        max_nodes=IntProp("max_nodes"),
        instance_family=EnumProp("instance_family", InstanceFamily),
        auto_resume=BoolProp("auto_resume"),
        initially_suspended=BoolProp("initially_suspended"),
        auto_suspend_secs=IntProp("auto_suspend_secs"),
        comment=StringProp("comment"),
    )
    scope = AccountScope()
    spec = _ComputePool

    def __init__(
        self,
        name: str,
        owner: str = "ACCOUNTADMIN",
        min_nodes: int = None,
        max_nodes: int = None,
        instance_family: InstanceFamily = None,
        auto_resume: bool = True,
        initially_suspended: bool = None,
        auto_suspend_secs: int = 3600,
        comment: str = None,
        **kwargs,
    ):
        super().__init__(name, **kwargs)
        self._data: _ComputePool = _ComputePool(
            name=self._name,
            owner=owner,
            min_nodes=min_nodes,
            max_nodes=max_nodes,
            instance_family=instance_family,
            auto_resume=auto_resume,
            initially_suspended=initially_suspended,
            auto_suspend_secs=auto_suspend_secs,
            comment=comment,
        )



import pytest

from titan.adapters import permifrost
from titan.enums import ResourceType
from titan.privs import DatabasePriv, WarehousePriv
from titan.resources import Grant, RoleGrant
from titan.resources.resource import ResourcePointer


@pytest.mark.skip("skipping due to pending deprecation")
@pytest.mark.requires_snowflake
def test_permifrost(cursor):
    resources = permifrost.read_permifrost_config(cursor.connection, "tests/fixtures/adapters/permifrost.yml")
    assert ResourcePointer(name="loading", resource_type=ResourceType.WAREHOUSE) in resources
    assert Grant(priv=WarehousePriv.OPERATE, on_warehouse="loading", to="accountadmin") in resources
    assert RoleGrant(role="engineer", to_role="sysadmin") in resources
    assert ResourcePointer(name="raw", resource_type=ResourceType.DATABASE) in resources
    assert Grant(priv=DatabasePriv.USAGE, on_database="raw", to="sysadmin") in resources
    assert ResourcePointer(name="raw", resource_type=ResourceType.DATABASE) in resources
    assert RoleGrant(role="sysadmin", to_user="eburke") in resources
    assert RoleGrant(role="eburke", to_user="eburke") in resources



from .account import Account
from .aggregation_policy import AggregationPolicy
from .alert import Alert
from .api_integration import APIIntegration
from .authentication_policy import AuthenticationPolicy
from .catalog_integration import GlueCatalogIntegration, ObjectStoreCatalogIntegration
from .column import Column
from .compute_pool import ComputePool
from .database import Database
from .dynamic_table import DynamicTable
from .event_table import EventTable
from .external_access_integration import ExternalAccessIntegration
from .external_function import ExternalFunction
from .failover_group import FailoverGroup
from .file_format import CSVFileFormat, JSONFileFormat, ParquetFileFormat
from .function import JavascriptUDF, PythonUDF
from .grant import FutureGrant, Grant, GrantOnAll, RoleGrant
from .hybrid_table import HybridTable
from .image_repository import ImageRepository
from .materialized_view import MaterializedView
from .network_policy import NetworkPolicy
from .network_rule import NetworkRule
from .notification_integration import (
    AWSOutboundNotificationIntegration,
    AzureInboundNotificationIntegration,
    AzureOutboundNotificationIntegration,
    EmailNotificationIntegration,
    GCPInboundNotificationIntegration,
    GCPOutboundNotificationIntegration,
)
from .notebook import Notebook
from .packages_policy import PackagesPolicy
from .password_policy import PasswordPolicy
from .pipe import Pipe
from .procedure import PythonStoredProcedure
from .replication_group import ReplicationGroup
from .resource import Resource
from .resource_monitor import ResourceMonitor
from .role import DatabaseRole, Role
from .schema import Schema
from .secret import GenericSecret, OAuthSecret, PasswordSecret
from .security_integration import (
    APIAuthenticationSecurityIntegration,
    SnowflakePartnerOAuthSecurityIntegration,
    SnowservicesOAuthSecurityIntegration,
)
from .sequence import Sequence
from .service import Service
from .share import Share
from .stage import ExternalStage, InternalStage
from .storage_integration import (
    AzureStorageIntegration,
    GCSStorageIntegration,
    S3StorageIntegration,
)
from .stream import StageStream, TableStream, ViewStream  # ExternalTableStream
from .table import Table  # , CreateTableAsSelect
from .tag import Tag, TagReference
from .task import Task
from .user import User
from .view import View
from .warehouse import Warehouse

__all__ = [
    "Account",
    "AggregationPolicy",
    "Alert",
    "APIAuthenticationSecurityIntegration",
    "APIIntegration",
    "AuthenticationPolicy",
    "AWSOutboundNotificationIntegration",
    "AzureInboundNotificationIntegration",
    "AzureOutboundNotificationIntegration",
    "AzureStorageIntegration",
    "Column",
    "ComputePool",
    # "CreateTableAsSelect",
    "CSVFileFormat",
    "Database",
    "DatabaseRole",
    "DynamicTable",
    "EmailNotificationIntegration",
    "EventTable",
    "ExternalAccessIntegration",
    "ExternalFunction",
    "ExternalStage",
    "FailoverGroup",
    "JSONFileFormat",
    "FutureGrant",
    "GCPInboundNotificationIntegration",
    "GCPOutboundNotificationIntegration",
    "GCSStorageIntegration",
    "GenericSecret",
    "GlueCatalogIntegration",
    "Grant",
    "GrantOnAll",
    "HybridTable",
    "ImageRepository",
    "InternalStage",
    "JavascriptUDF",
    "MaterializedView",
    "NetworkPolicy",
    "NetworkRule",
    "Notebook",
    "OAuthSecret",
    "ObjectStoreCatalogIntegration",
    "PackagesPolicy",
    "ParquetFileFormat",
    "PasswordPolicy",
    "PasswordSecret",
    "Pipe",
    "PythonStoredProcedure",
    "PythonUDF",
    "ReplicationGroup",
    "Resource",
    "ResourceMonitor",
    "Role",
    "RoleGrant",
    "S3StorageIntegration",
    "Schema",
    "Sequence",
    "Service",
    "Share",
    "SnowflakePartnerOAuthSecurityIntegration",
    "SnowservicesOAuthSecurityIntegration",
    "StageStream",
    "Table",
    "TableStream",
    "Tag",
    "TagReference",
    "Task",
    "User",
    "View",
    "ViewStream",
    "Warehouse",
]



import pytest

from titan import resources as res
from titan.blueprint import Action, Blueprint, InvalidOwnerException, MissingPrivilegeException, compile_plan_to_sql
from titan.identifiers import parse_URN
from titan.privs import AccountPriv, GrantedPrivilege
from titan.resource_name import ResourceName
from titan.resources.tag import tag_reference_for_resource


@pytest.fixture
def session_ctx() -> dict:
    return {
        "account": "SOMEACCT",
        "account_locator": "ABCD123",
        "role": "SYSADMIN",
        "available_roles": [
            "SYSADMIN",
            "USERADMIN",
            "ACCOUNTADMIN",
            "SECURITYADMIN",
            "PUBLIC",
        ],
        "role_privileges": {},
    }


@pytest.fixture
def remote_state() -> dict:
    return {
        parse_URN("urn::ABCD123:account/SOMEACCT"): {},
    }


def test_default_owner(session_ctx, remote_state):
    warehouse = res.Warehouse(name="test_warehouse")
    assert warehouse._data.owner.name == "SYSADMIN"
    blueprint = Blueprint(resources=[warehouse])
    manifest = blueprint.generate_manifest(session_ctx)
    plan = blueprint._plan(remote_state, manifest)
    assert len(plan) == 1
    assert plan[0].action == Action.ADD
    assert plan[0].after["owner"] == "SYSADMIN"


def test_non_default_owner(session_ctx, remote_state):
    warehouse = res.Warehouse(name="test_warehouse", owner="ACCOUNTADMIN")
    assert warehouse._data.owner.name == "ACCOUNTADMIN"
    blueprint = Blueprint(resources=[warehouse])
    manifest = blueprint.generate_manifest(session_ctx)
    plan = blueprint._plan(remote_state, manifest)
    assert len(plan) == 1
    assert plan[0].action == Action.ADD
    assert plan[0].after["owner"] == "ACCOUNTADMIN"


def test_custom_role_owner(session_ctx, remote_state):
    role = res.Role(name="CUSTOMROLE")
    grant = res.RoleGrant(role=role, to_role="SYSADMIN")
    warehouse = res.Warehouse(name="test_warehouse", owner=role)
    assert warehouse._data.owner.name == "CUSTOMROLE"
    blueprint = Blueprint(resources=[role, grant, warehouse])
    manifest = blueprint.generate_manifest(session_ctx)
    plan = blueprint._plan(remote_state, manifest)
    assert len(plan) == 3
    assert plan[0].action == Action.ADD
    assert plan[0].urn == parse_URN("urn::ABCD123:role/CUSTOMROLE")
    assert plan[1].action == Action.ADD
    assert plan[1].urn == parse_URN("urn::ABCD123:role_grant/CUSTOMROLE?role=SYSADMIN")
    assert plan[2].action == Action.ADD
    assert plan[2].urn == parse_URN("urn::ABCD123:warehouse/test_warehouse")
    assert plan[2].after["owner"] == "CUSTOMROLE"


def test_invalid_custom_role_owner(session_ctx):
    role = res.Role(name="INVALIDROLE")
    warehouse = res.Warehouse(name="test_warehouse", owner=role)
    blueprint = Blueprint(resources=[role, warehouse])
    with pytest.raises(InvalidOwnerException):
        blueprint.generate_manifest(session_ctx)


def test_transfer_ownership(session_ctx, remote_state):
    remote_state = remote_state.copy()
    remote_state[parse_URN("urn::ABCD123:role/test_role")] = {
        "name": "test_role",
        "owner": "ACCOUNTADMIN",
        "comment": None,
    }

    role = res.Role(name="test_role")
    blueprint = Blueprint(resources=[role])
    manifest = blueprint.generate_manifest(session_ctx)
    plan = blueprint._plan(remote_state, manifest)
    assert len(plan) == 1
    assert plan[0].action == Action.TRANSFER
    assert plan[0].after["owner"] == "USERADMIN"


def test_transfer_ownership_with_changes(session_ctx, remote_state):
    remote_state = remote_state.copy()
    remote_state[parse_URN("urn::ABCD123:role/test_role")] = {
        "name": "test_role",
        "owner": "ACCOUNTADMIN",
        "comment": None,
    }

    role = res.Role(name="test_role", comment="This comment has been added")
    blueprint = Blueprint(resources=[role])
    manifest = blueprint.generate_manifest(session_ctx)
    plan = blueprint._plan(remote_state, manifest)
    assert len(plan) == 2
    assert plan[0].action == Action.CHANGE
    assert plan[0].after["comment"] == "This comment has been added"
    assert plan[1].action == Action.TRANSFER
    assert plan[1].after["owner"] == "USERADMIN"


def test_resource_has_custom_role_owner_with_create_priv(session_ctx, remote_state):
    session_ctx = session_ctx.copy()
    session_ctx["available_roles"].append(ResourceName("test_role"))
    session_ctx["role_privileges"] = {
        "TEST_ROLE": [
            GrantedPrivilege(privilege=AccountPriv.CREATE_WAREHOUSE, on="ABCD123"),
        ]
    }

    warehouse = res.Warehouse(name="test_warehouse", owner="test_role")
    blueprint = Blueprint(resources=[warehouse])
    manifest = blueprint.generate_manifest(session_ctx)
    plan = blueprint._plan(remote_state, manifest)
    assert len(plan) == 1
    assert plan[0].action == Action.ADD
    assert plan[0].urn == parse_URN("urn::ABCD123:warehouse/test_warehouse")

    sql_commands = compile_plan_to_sql(session_ctx, plan)
    assert sql_commands[0] == "USE SECONDARY ROLES ALL"
    assert sql_commands[1] == "USE ROLE TEST_ROLE"
    assert sql_commands[2].startswith("CREATE WAREHOUSE TEST_WAREHOUSE")


def test_resource_is_transferred_to_custom_role_owner(session_ctx, remote_state):
    session_ctx = session_ctx.copy()
    session_ctx["available_roles"].append(ResourceName("test_role"))

    warehouse = res.Warehouse(name="test_warehouse", owner="test_role")
    blueprint = Blueprint(resources=[warehouse])
    manifest = blueprint.generate_manifest(session_ctx)
    plan = blueprint._plan(remote_state, manifest)
    assert len(plan) == 1
    assert plan[0].action == Action.ADD
    assert plan[0].urn == parse_URN("urn::ABCD123:warehouse/test_warehouse")

    sql_commands = compile_plan_to_sql(session_ctx, plan)
    assert sql_commands[0] == "USE SECONDARY ROLES ALL"
    assert sql_commands[1] == "USE ROLE ACCOUNTADMIN"
    assert sql_commands[2].startswith("CREATE WAREHOUSE TEST_WAREHOUSE")
    assert sql_commands[3] == "GRANT OWNERSHIP ON WAREHOUSE TEST_WAREHOUSE TO ROLE TEST_ROLE COPY CURRENT GRANTS"


def test_resource_cant_be_created(remote_state):
    session_ctx = {
        "account": "SOMEACCT",
        "account_locator": "ABCD123",
        "role": "TEST_ROLE",
        "available_roles": [
            "TEST_ROLE",
        ],
        "role_privileges": {},
    }
    warehouse = res.Warehouse(name="test_warehouse", owner="test_role")
    blueprint = Blueprint(resources=[warehouse])
    manifest = blueprint.generate_manifest(session_ctx)
    plan = blueprint._plan(remote_state, manifest)
    assert len(plan) == 1
    assert plan[0].action == Action.ADD
    assert plan[0].urn == parse_URN("urn::ABCD123:warehouse/test_warehouse")

    with pytest.raises(MissingPrivilegeException):
        compile_plan_to_sql(session_ctx, plan)


def test_grant_with_grant_admin_custom_role(remote_state):
    session_ctx = {
        "account": "SOMEACCT",
        "account_locator": "ABCD123",
        "role": "GRANT_ADMIN",
        "available_roles": [
            "GRANT_ADMIN",
        ],
        "role_privileges": {
            "GRANT_ADMIN": [
                GrantedPrivilege(privilege=AccountPriv.MANAGE_GRANTS, on="ABCD123"),
            ]
        },
    }

    grant = res.RoleGrant(role="GRANT_ADMIN", to_role="SYSADMIN")
    blueprint = Blueprint(resources=[grant])
    manifest = blueprint.generate_manifest(session_ctx)
    plan = blueprint._plan(remote_state, manifest)
    assert len(plan) == 1
    assert plan[0].action == Action.ADD
    assert plan[0].urn == parse_URN("urn::ABCD123:role_grant/GRANT_ADMIN?role=SYSADMIN")
    compile_plan_to_sql(session_ctx, plan)


def test_tag_reference_with_tag_admin_custom_role():
    session_ctx = {
        "account": "SOMEACCT",
        "account_locator": "ABCD123",
        "role": "TAG_ADMIN",
        "available_roles": [
            "TAG_ADMIN",
        ],
        "role_privileges": {
            "TAG_ADMIN": [
                GrantedPrivilege(privilege=AccountPriv.APPLY_TAG, on="ABCD123"),
            ]
        },
        "tags": ["tags.tags.cost_center"],
    }

    remote_state = {
        parse_URN("urn::ABCD123:account/SOMEACCT"): {},
    }

    tag_reference = res.TagReference(
        object_name="SOME_ROLE",
        object_domain="ROLE",
        tags={"tags.tags.cost_center": "finance"},
    )
    blueprint = Blueprint(resources=[tag_reference])
    manifest = blueprint.generate_manifest(session_ctx)
    plan = blueprint._plan(remote_state, manifest)
    assert len(plan) == 1
    assert plan[0].action == Action.ADD
    assert plan[0].urn == parse_URN("urn::ABCD123:tag_reference/SOME_ROLE?domain=ROLE")
    sql_commands = compile_plan_to_sql(session_ctx, plan)
    assert len(sql_commands) == 3
    assert sql_commands[0] == "USE SECONDARY ROLES ALL"
    assert sql_commands[1] == "USE ROLE TAG_ADMIN"
    assert sql_commands[2] == "ALTER ROLE SOME_ROLE SET TAG tags.tags.cost_center='finance'"



from dataclasses import dataclass

from .resource import Resource, ResourceSpec, NamedResource
from .role import Role
from ..enums import BinaryFormat, Compression, FileType, ResourceType
from ..props import (
    BoolProp,
    EnumProp,
    IntProp,
    Props,
    StringListProp,
    StringProp,
)
from ..resource_name import ResourceName
from ..scope import SchemaScope


@dataclass(unsafe_hash=True)
class _CSVFileFormat(ResourceSpec):
    name: ResourceName
    owner: Role = "SYSADMIN"
    type: FileType = "CSV"
    compression: Compression = "AUTO"
    record_delimiter: str = "\n"
    field_delimiter: str = ","
    file_extension: str = None
    parse_header: bool = False
    skip_header: int = 0
    skip_blank_lines: bool = False
    date_format: str = "AUTO"
    time_format: str = "AUTO"
    timestamp_format: str = "AUTO"
    binary_format: BinaryFormat = "HEX"
    escape: str = None
    escape_unenclosed_field: str = "\\"
    trim_space: bool = False
    field_optionally_enclosed_by: str = None
    null_if: list[str] = None
    error_on_column_count_mismatch: bool = True
    replace_invalid_characters: bool = False
    empty_field_as_null: bool = True
    skip_byte_order_mark: bool = True
    encoding: str = "UTF8"
    comment: str = None

    def __post_init__(self):
        super().__post_init__()
        if self.null_if is None:
            self.null_if = ["\n"]


class CSVFileFormat(NamedResource, Resource):
    """
    Description:
        Defines the specifications for a CSV file format in Snowflake, including delimiters, encoding, and compression options.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-file-format

    Fields:
        compression (string or Compression): The compression type used for the file format.
        record_delimiter (string): Specifies the character that delimits records. Defaults to "\n".
        field_delimiter (string): Specifies the character that delimits fields.
        file_extension (string): The file extension used for files of this format.
        parse_header (bool): Whether to parse the first line of the file as a header. Defaults to False.
        skip_header (int): The number of header lines to skip before parsing.
        skip_blank_lines (bool): Whether to skip over blank lines. Defaults to False.
        date_format (string): The format used for date values. Defaults to "AUTO".
        time_format (string): The format used for time values. Defaults to "AUTO".
        timestamp_format (string): The format used for timestamp values. Defaults to "AUTO".
        binary_format (string or BinaryFormat): The format used for binary data. Defaults to HEX.
        escape (string): The escape character used in the file format.
        escape_unenclosed_field (string): The escape character for unenclosed fields. Defaults to "\\".
        trim_space (bool): Whether to trim spaces from fields. Defaults to False.
        field_optionally_enclosed_by (string): A character that may optionally enclose fields.
        null_if (list): A list of string values that should be interpreted as NULL.
        error_on_column_count_mismatch (bool, required): Whether to raise an error on column count mismatch. Defaults to True.
        replace_invalid_characters (bool): Whether to replace invalid characters. Defaults to False.
        empty_field_as_null (bool): Whether to treat empty fields as NULL.
        skip_byte_order_mark (bool): Whether to skip the byte order mark. Defaults to True.
        encoding (string): The file encoding. Defaults to "UTF8".
        comment (string): A comment regarding the file format.

    Python:

        ```python
        csv_file_format = CSVFileFormat(
            name="some_csv_file_format",
            compression=Compression.GZIP,
            field_delimiter=',',
            file_extension='csv',
            parse_header=True
        )
        ```

    Yaml:

        ```yaml
        csv_file_formats:
          - name: some_csv_file_format
            compression: GZIP
            field_delimiter: ','
            file_extension: 'csv'
            parse_header: true
        ```
    """

    resource_type = ResourceType.FILE_FORMAT
    props = Props(
        type=EnumProp("type", [FileType.CSV]),
        compression=EnumProp("compression", Compression),
        record_delimiter=StringProp("record_delimiter", alt_tokens=["NONE"]),
        field_delimiter=StringProp("field_delimiter", alt_tokens=["NONE"]),
        file_extension=StringProp("file_extension"),
        parse_header=BoolProp("parse_header"),
        skip_header=IntProp("skip_header"),
        skip_blank_lines=BoolProp("skip_blank_lines"),
        date_format=StringProp("date_format", alt_tokens=["AUTO"]),
        time_format=StringProp("time_format", alt_tokens=["AUTO"]),
        timestamp_format=StringProp("timestamp_format", alt_tokens=["AUTO"]),
        binary_format=EnumProp("binary_format", BinaryFormat),
        escape=StringProp("escape", alt_tokens=["NONE"]),
        escape_unenclosed_field=StringProp("escape_unenclosed_field", alt_tokens=["NONE"]),
        trim_space=BoolProp("trim_space"),
        field_optionally_enclosed_by=StringProp("field_optionally_enclosed_by", alt_tokens=["NONE"]),
        null_if=StringListProp("null_if", parens=True),
        error_on_column_count_mismatch=BoolProp("error_on_column_count_mismatch"),
        replace_invalid_characters=BoolProp("replace_invalid_characters"),
        empty_field_as_null=BoolProp("empty_field_as_null"),
        skip_byte_order_mark=BoolProp("skip_byte_order_mark"),
        encoding=StringProp("encoding", alt_tokens=["UTF8"]),
        comment=StringProp("comment"),
    )
    scope = SchemaScope()
    spec = _CSVFileFormat

    def __init__(
        self,
        name: str,
        owner: str = "SYSADMIN",
        compression: Compression = "AUTO",
        record_delimiter: str = "\n",
        field_delimiter: str = ",",
        file_extension: str = None,
        parse_header: bool = False,
        skip_header: int = 0,
        skip_blank_lines: bool = False,
        date_format: str = "AUTO",
        time_format: str = "AUTO",
        timestamp_format: str = "AUTO",
        binary_format: BinaryFormat = "HEX",
        escape: str = None,
        escape_unenclosed_field: str = "\\",
        trim_space: bool = False,
        field_optionally_enclosed_by: str = None,
        null_if: list[str] = None,
        error_on_column_count_mismatch: bool = True,
        replace_invalid_characters: bool = False,
        empty_field_as_null: bool = True,
        skip_byte_order_mark: bool = True,
        encoding: str = "UTF8",
        comment: str = None,
        **kwargs,
    ):
        kwargs.pop("type", None)
        super().__init__(name, **kwargs)
        self._data: _CSVFileFormat = _CSVFileFormat(
            name=self._name,
            owner=owner,
            compression=compression,
            record_delimiter=record_delimiter,
            field_delimiter=field_delimiter,
            file_extension=file_extension,
            parse_header=parse_header,
            skip_header=skip_header,
            skip_blank_lines=skip_blank_lines,
            date_format=date_format,
            time_format=time_format,
            timestamp_format=timestamp_format,
            binary_format=binary_format,
            escape=escape,
            escape_unenclosed_field=escape_unenclosed_field,
            trim_space=trim_space,
            field_optionally_enclosed_by=field_optionally_enclosed_by,
            null_if=null_if,
            error_on_column_count_mismatch=error_on_column_count_mismatch,
            replace_invalid_characters=replace_invalid_characters,
            empty_field_as_null=empty_field_as_null,
            skip_byte_order_mark=skip_byte_order_mark,
            encoding=encoding,
            comment=comment,
        )


@dataclass(unsafe_hash=True)
class _JSONFileFormat(ResourceSpec):
    name: ResourceName
    owner: Role = "SYSADMIN"
    type: FileType = FileType.JSON
    compression: Compression = Compression.AUTO
    date_format: str = "AUTO"
    time_format: str = "AUTO"
    timestamp_format: str = "AUTO"
    binary_format: BinaryFormat = BinaryFormat.HEX
    trim_space: bool = False
    null_if: list[str] = None
    file_extension: str = None
    enable_octal: bool = False
    allow_duplicate: bool = False
    strip_outer_array: bool = False
    strip_null_values: bool = False
    replace_invalid_characters: bool = False
    ignore_utf8_errors: bool = False
    skip_byte_order_mark: bool = True
    comment: str = None

    def __post_init__(self):
        super().__post_init__()
        if self.null_if is None:
            self.null_if = []


class JSONFileFormat(NamedResource, Resource):
    """
    Description:
        A JSON file format in Snowflake.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-file-format

    Fields:
        name (string, required): The name of the file format.
        owner (string or Role): The owner role of the file format. Defaults to "SYSADMIN".
        compression (string): The compression type for the file format. Defaults to "AUTO".
        date_format (string): The format used for date values. Defaults to "AUTO".
        time_format (string): The format used for time values. Defaults to "AUTO".
        timestamp_format (string): The format used for timestamp values. Defaults to "AUTO".
        binary_format (BinaryFormat): The format used for binary data. Defaults to HEX.
        trim_space (bool): Whether to trim spaces. Defaults to False.
        null_if (list): A list of strings to be interpreted as NULL.
        file_extension (string): The file extension used for files of this format.
        enable_octal (bool): Whether to enable octal values. Defaults to False.
        allow_duplicate (bool): Whether to allow duplicate keys. Defaults to False.
        strip_outer_array (bool): Whether to strip the outer array. Defaults to False.
        strip_null_values (bool): Whether to strip null values. Defaults to False.
        replace_invalid_characters (bool): Whether to replace invalid characters. Defaults to False.
        ignore_utf8_errors (bool): Whether to ignore UTF-8 errors. Defaults to False.
        skip_byte_order_mark (bool): Whether to skip the byte order mark. Defaults to True.
        comment (string): A comment for the file format.

    Python:

        ```python
        file_format = JSONFileFormat(
            name="some_json_file_format",
            owner="SYSADMIN",
            compression="AUTO",
            date_format="AUTO",
            time_format="AUTO",
            timestamp_format="AUTO",
            binary_format=BinaryFormat.HEX,
            trim_space=False,
            null_if=["NULL"],
            file_extension="json",
            enable_octal=False,
            allow_duplicate=False,
            strip_outer_array=False,
            strip_null_values=False,
            replace_invalid_characters=False,
            ignore_utf8_errors=False,
            skip_byte_order_mark=True,
            comment="This is a JSON file format."
        )
        ```

    Yaml:

        ```yaml
        file_formats:
          - name: some_json_file_format
            owner: SYSADMIN
            compression: AUTO
            date_format: AUTO
            time_format: AUTO
            timestamp_format: AUTO
            binary_format: HEX
            trim_space: false
            null_if:
              - NULL
            file_extension: json
            enable_octal: false
            allow_duplicate: false
            strip_outer_array: false
            strip_null_values: false
            replace_invalid_characters: false
            ignore_utf8_errors: false
            skip_byte_order_mark: true
            comment: This is a JSON file format.
        ```
    """

    resource_type = ResourceType.FILE_FORMAT
    props = Props(
        type=EnumProp("type", [FileType.JSON]),
        compression=EnumProp("compression", Compression),
        date_format=StringProp("date_format", alt_tokens=["AUTO"]),
        time_format=StringProp("time_format", alt_tokens=["AUTO"]),
        timestamp_format=StringProp("timestamp_format", alt_tokens=["AUTO"]),
        binary_format=EnumProp("binary_format", BinaryFormat),
        trim_space=BoolProp("trim_space"),
        null_if=StringListProp("null_if", parens=True),
        file_extension=StringProp("file_extension"),
        enable_octal=BoolProp("enable_octal"),
        allow_duplicate=BoolProp("allow_duplicate"),
        strip_outer_array=BoolProp("strip_outer_array"),
        strip_null_values=BoolProp("strip_null_values"),
        replace_invalid_characters=BoolProp("replace_invalid_characters"),
        ignore_utf8_errors=BoolProp("ignore_utf8_errors"),
        skip_byte_order_mark=BoolProp("skip_byte_order_mark"),
        comment=StringProp("comment"),
    )
    scope = SchemaScope()
    spec = _JSONFileFormat

    def __init__(
        self,
        name: str,
        owner: str = "SYSADMIN",
        compression: Compression = Compression.AUTO,
        date_format: str = "AUTO",
        time_format: str = "AUTO",
        timestamp_format: str = "AUTO",
        binary_format: BinaryFormat = BinaryFormat.HEX,
        trim_space: bool = False,
        null_if: list[str] = None,
        file_extension: str = "json",
        enable_octal: bool = False,
        allow_duplicate: bool = False,
        strip_outer_array: bool = False,
        strip_null_values: bool = False,
        replace_invalid_characters: bool = False,
        ignore_utf8_errors: bool = False,
        skip_byte_order_mark: bool = True,
        comment: str = None,
        **kwargs,
    ):
        kwargs.pop("type", None)
        super().__init__(name, **kwargs)
        self._data: _JSONFileFormat = _JSONFileFormat(
            name=self._name,
            owner=owner,
            compression=compression,
            date_format=date_format,
            time_format=time_format,
            timestamp_format=timestamp_format,
            binary_format=binary_format,
            trim_space=trim_space,
            null_if=null_if,
            file_extension=file_extension,
            enable_octal=enable_octal,
            allow_duplicate=allow_duplicate,
            strip_outer_array=strip_outer_array,
            strip_null_values=strip_null_values,
            replace_invalid_characters=replace_invalid_characters,
            ignore_utf8_errors=ignore_utf8_errors,
            skip_byte_order_mark=skip_byte_order_mark,
            comment=comment,
        )


@dataclass(unsafe_hash=True)
class _ParquetFileFormat(ResourceSpec):
    name: ResourceName
    owner: Role = "SYSADMIN"
    type: FileType = FileType.PARQUET
    compression: Compression = Compression.AUTO
    binary_as_text: bool = True
    trim_space: bool = False
    replace_invalid_characters: bool = False
    null_if: list[str] = None
    comment: str = None

    def __post_init__(self):
        super().__post_init__()
        if self.null_if is None:
            self.null_if = []


class ParquetFileFormat(NamedResource, Resource):
    """
    Description:
        A Parquet file format in Snowflake.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-file-format

    Fields:
        name (string, required): The name of the file format.
        owner (string or Role): The owner role of the file format. Defaults to "SYSADMIN".
        compression (string): The compression type for the file format. Defaults to "AUTO".
        binary_as_text (bool): Whether to interpret binary data as text. Defaults to True.
        trim_space (bool): Whether to trim spaces. Defaults to False.
        replace_invalid_characters (bool): Whether to replace invalid characters. Defaults to False.
        null_if (list): A list of strings to be interpreted as NULL.
        comment (string): A comment for the file format.

    Python:

        ```python
        file_format = ParquetFileFormat(
            name="some_file_format",
            owner="SYSADMIN",
            compression="AUTO",
            binary_as_text=True,
            trim_space=False,
            replace_invalid_characters=False,
            null_if=["NULL"],
            comment="This is a Parquet file format."
        )
        ```

    Yaml:

        ```yaml
        file_formats:
          - name: some_file_format
            owner: SYSADMIN
            compression: AUTO
            binary_as_text: true
            trim_space: false
            replace_invalid_characters: false
            null_if:
              - NULL
            comment: This is a Parquet file format.
        ```
    """

    resource_type = ResourceType.FILE_FORMAT
    props = Props(
        type=EnumProp("type", [FileType.PARQUET]),
        compression=EnumProp("compression", Compression),
        binary_as_text=BoolProp("binary_as_text"),
        trim_space=BoolProp("trim_space"),
        replace_invalid_characters=BoolProp("replace_invalid_characters"),
        null_if=StringListProp("null_if", parens=True),
        comment=StringProp("comment"),
    )
    scope = SchemaScope()
    spec = _ParquetFileFormat

    def __init__(
        self,
        name: str,
        owner: str = "SYSADMIN",
        compression: str = "AUTO",
        binary_as_text: bool = True,
        trim_space: bool = False,
        replace_invalid_characters: bool = False,
        null_if: list[str] = None,
        comment: str = None,
        **kwargs,
    ):
        kwargs.pop("type", None)
        super().__init__(name, **kwargs)
        self._data: _ParquetFileFormat = _ParquetFileFormat(
            name=self._name,
            owner=owner,
            compression=compression,
            binary_as_text=binary_as_text,
            trim_space=trim_space,
            replace_invalid_characters=replace_invalid_characters,
            null_if=null_if,
            comment=comment,
        )


FileFormatMap = {
    FileType.CSV: CSVFileFormat,
    FileType.JSON: JSONFileFormat,
    # FileType.AVRO: AvroFileFormat,
    # FileType.ORC: OrcFileFormat,
    FileType.PARQUET: ParquetFileFormat,
    # FileType.XML: XMLFileFormat,
}


def _resolver(data: dict):
    return FileFormatMap[FileType(data["type"])]


Resource.__resolvers__[ResourceType.FILE_FORMAT] = _resolver



import copy

import pytest

from dataclasses import fields
from typing import get_args, get_origin

from tests.helpers import get_json_fixtures

from titan.resources import Resource
from titan.resource_name import ResourceName

JSON_FIXTURES = list(get_json_fixtures())


@pytest.fixture(
    params=JSON_FIXTURES,
    ids=[resource_cls.__name__ for resource_cls, _ in JSON_FIXTURES],
    scope="function",
)
def resource(request):
    resource_cls, data = request.param
    yield resource_cls, data


def _resource_field_type_is_resource(field):
    if isinstance(field.type, str) and field.name == "owner" and field.type == "Role":
        return True
    elif issubclass(field.type, Resource):
        return True
    elif get_origin(field.type) == list and issubclass(get_args(field.type)[0], Resource):
        return True

    return False


def _resource_names_are_eq(lhs, rhs):
    if lhs is None and rhs is None:
        return True
    if isinstance(lhs, list):
        return [ResourceName(item) for item in lhs] == [ResourceName(item) for item in rhs]
    else:
        return ResourceName(lhs) == ResourceName(rhs)


def test_data_identity(resource):
    resource_cls, data = resource
    data: dict = copy.deepcopy(data)
    instance = resource_cls(**data)

    serialized: dict = instance.to_dict()
    if "name" in serialized:
        assert _resource_names_are_eq(serialized.pop("name"), data.pop("name"))
    if "columns" in serialized:
        lhs_cols = serialized.pop("columns", []) or []
        rhs_cols = data.pop("columns", []) or []
        assert len(lhs_cols) == len(rhs_cols)
        for lhs, rhs in zip(lhs_cols, rhs_cols):
            if "name" in lhs:
                assert _resource_names_are_eq(lhs.pop("name"), rhs.pop("name"))
            assert lhs == rhs
    if "args" in serialized:
        lhs_args = serialized.pop("args", []) or []
        rhs_args = data.pop("args", []) or []
        assert len(lhs_args) == len(rhs_args)
        for lhs, rhs in zip(lhs_args, rhs_args):
            if "name" in lhs:
                assert _resource_names_are_eq(lhs.pop("name"), rhs.pop("name"))
            assert lhs == rhs
    for field in fields(instance._data):
        if field.name in ["name", "columns", "args"]:
            continue
        if _resource_field_type_is_resource(field):
            assert _resource_names_are_eq(serialized.pop(field.name), data.pop(field.name))

    assert serialized == data


def test_sql_identity(resource):
    resource_cls, data = resource
    instance = resource_cls(**data)
    sql = instance.create_sql()
    new = resource_cls.from_sql(sql)
    new_dict = new.to_dict()
    instance_dict = instance.to_dict()
    if "name" in new_dict:
        assert ResourceName(new_dict.pop("name")) == ResourceName(instance_dict.pop("name"))

    assert new_dict == instance_dict



import json
import sys
from abc import ABC
from typing import Dict, Optional

import pyparsing as pp

from .builder import tidy_sql
from .enums import DataType
from .parse import (
    ANY,
    ARROW,
    EQUALS,
    FullyQualifiedIdentifier,
    Identifier,
    Keyword,
    Keywords,
    Literals,
    _in_parens,
    _parse_props,
    _parser_has_results_name,
)

__this__ = sys.modules[__name__]


class Prop(ABC):
    """
    A Prop is a named expression that can be parsed from a SQL string.
    """

    # TODO: find a better home for alt_tokens
    def __init__(self, label, value_expr=ANY(), eq=True, parens=False, alt_tokens=[], consume=[]):
        self.label = label
        self.eq = eq
        self.parens = parens
        self.alt_tokens = set([tok.lower() for tok in alt_tokens])

        if isinstance(consume, str):
            consume = [consume]
        consume_expr = None
        if consume:
            consume_expr = pp.And([pp.Opt(Keyword(tok)) for tok in consume]).suppress()

        label_expr = None
        if self.label:
            label_expr = Keywords(self.label).suppress()

        eq_expr = None
        if self.eq:
            eq_expr = EQUALS()

        if not _parser_has_results_name(value_expr, "prop_value"):
            value_expr = value_expr("prop_value")

        if parens:
            value_expr = _in_parens(value_expr)

        expressions = []
        for expr in [consume_expr, label_expr, consume_expr, eq_expr, value_expr]:
            if expr:
                expressions.append(expr)

        self.parser = pp.And(expressions)

    def __repr__(self):  # pragma: no cover
        return f"{self.__class__.__name__}('{self.label}')"

    def parse(self, sql):
        parsed = self.parser.parse_string(sql)
        prop_value = parsed["prop_value"]
        if isinstance(prop_value, pp.ParseResults):
            prop_value = prop_value.as_list()
        return self.typecheck(prop_value)

    def typecheck(self, prop_value):
        raise NotImplementedError

    def render(self, value):
        raise NotImplementedError


class Props:
    def __init__(self, _name: Optional[str] = None, _start_token: Optional[str] = None, **props: Prop):
        self.props: Dict[str, Prop] = props
        self.name = _name
        self.start_token = Literals(_start_token) if _start_token else None

    def __repr__(self):
        return f"Props(num:{len(self.props)})"

    def __getitem__(self, key: str) -> Prop:
        return self.props[key]

    def to_json(self):
        return json.dumps(self, default=lambda obj: obj.__dict__)

    def render(self, data):
        rendered = []
        for prop_kwarg, prop in self.props.items():
            value = data.get(prop_kwarg)
            if value is None:
                continue
            rendered.append(prop.render(value))
        return tidy_sql(rendered)


class BoolProp(Prop):
    def typecheck(self, prop_value):
        if prop_value.lower() not in ["true", "false"]:
            raise ValueError(f"Invalid boolean value: {prop_value}")
        return prop_value.lower() == "true"

    def render(self, value):
        if value is None:
            return ""
        return tidy_sql(
            self.label.upper(),
            "=" if self.eq else "",
            str(value).upper(),
        )


class IntProp(Prop):
    def typecheck(self, prop_value):
        try:
            return int(prop_value)
        except ValueError:
            raise ValueError(f"Invalid integer value: {prop_value}")

    def render(self, value):
        if value is None:
            return ""
        return tidy_sql(
            self.label.upper(),
            "=" if self.eq else "",
            value,
        )


class StringProp(Prop):
    def typecheck(self, prop_value):
        return prop_value

    def render(self, value):
        if value is None:
            return ""
        return tidy_sql(
            self.label.upper(),
            "=" if self.eq else "",
            f"$${value}$$",
        )


class FlagProp(Prop):
    def __init__(self, label):
        super().__init__(label, eq=False)
        self.parser = Keywords(self.label)("prop_value")

    def typecheck(self, _):
        return True

    def render(self, value):
        return self.label.upper() if value else ""


class IdentifierProp(Prop):
    def __init__(self, label, **kwargs):
        super().__init__(label, value_expr=FullyQualifiedIdentifier(), **kwargs)

    def typecheck(self, prop_value):
        return ".".join(prop_value)

    def render(self, value):
        if value is None:
            return ""
        if not isinstance(value, str) and hasattr(value, "name"):
            value = value.name
        return tidy_sql(
            self.label.upper(),
            "=" if self.eq else "",
            value,
        )


class IdentifierListProp(Prop):
    def __init__(self, label, **kwargs):
        value_expr = pp.delimited_list(pp.Group(FullyQualifiedIdentifier()))
        super().__init__(label, value_expr=value_expr, **kwargs)

    def typecheck(self, prop_values):
        return [".".join(id_parts) for id_parts in prop_values]

    def render(self, values):
        if values is None:
            return ""
        value_list = ", ".join(map(str, values))
        if self.parens:
            value_list = f"({value_list})"
        return tidy_sql(
            self.label.upper(),
            "=" if self.eq else "",
            value_list,
        )


class StringListProp(Prop):
    def __init__(self, label, **kwargs):
        value_expr = pp.delimited_list(ANY())
        super().__init__(label, value_expr=value_expr, **kwargs)

    def typecheck(self, prop_value):
        return [tok.strip(" ") for tok in prop_value]

    def render(self, values):
        if values is None or len(values) == 0:
            return ""
        value_list = ", ".join([f"$${v}$$" for v in values])
        return tidy_sql(
            self.label.upper(),
            "=" if self.eq else "",
            f"({value_list})" if self.parens else value_list,
        )


class PropSet(Prop):
    def __init__(self, label, props: Props):
        value_expr = pp.original_text_for(pp.nested_expr())
        super().__init__(label, value_expr)
        self.props: Props = props

    def typecheck(self, prop_value):
        prop_value = prop_value.strip("()")
        return _parse_props(self.props, prop_value)

    def render(self, values):
        if values is None or len(values) == 0:
            return ""
        eq = " = " if self.eq else " "
        value_str = self.props.render(values)
        value_str = f"({value_str})"
        return f"{self.label}{eq}{value_str}"


class TagsProp(Prop):
    """
    [ [ WITH ] TAG ( <tag_name> = '<tag_value>' [ , <tag_name> = '<tag_value>' , ... ] ) ]
    """

    def __init__(self):
        label = "TAG"
        value_expr = pp.delimited_list(ANY() + EQUALS() + ANY())
        super().__init__(label, value_expr=value_expr, eq=False, parens=True, consume="WITH")

    def typecheck(self, prop_value: list) -> dict:
        pairs = iter(prop_value)
        tags = {}
        for key in pairs:
            tags[key] = next(pairs)  # .strip("'")
        return tags

    def render(self, value: dict) -> str:
        if value is None:
            return ""
        tag_kv_pairs = ", ".join([f"{key} = '{value}'" for key, value in value.items()])
        return f"{self.label} ({tag_kv_pairs})"


class DictProp(Prop):
    """
    HEADERS = ( '<header_1>' = '<value_1>' [ , '<header_2>' = '<value_2>' ... ] )
    """

    def __init__(self, label, **kwargs):
        value_expr = pp.delimited_list(ANY() + EQUALS() + ANY())
        super().__init__(label, value_expr, **kwargs)

    def typecheck(self, prop_value):
        pairs = iter(prop_value)
        values = {}
        for key in pairs:
            values[key] = next(pairs)
        return values

    def render(self, value: dict) -> str:
        if value is None:
            return ""
        kv_pairs = ", ".join([f"'{key}' = '{value}'" for key, value in value.items()])
        eq = " = " if self.eq else " "
        return f"{self.label}{eq}({kv_pairs})"


class ReturnsProp(Prop):
    """
    RETURNS VARIANT
    RETURNS VARCHAR NOT NULL
    RETURNS NUMBER(38,0)
    RETURNS TABLE (event_date DATE, city VARCHAR, temperature NUMBER)
    """

    def __init__(self, label, **kwargs):
        data_type = pp.MatchFirst([Keywords(val.value) for val in set(DataType)]) | Keyword("TABLE")
        value_expr = pp.delimited_list(data_type + pp.Optional(pp.original_text_for(pp.nested_expr())))
        super().__init__(label, value_expr, **kwargs)

    def typecheck(self, prop_value):
        return "".join(prop_value)
        # if isinstance(prop_value, list):
        #     if len(prop_value) == 1:
        #         return {"data_type": DataType(prop_value[0])}
        #     else:
        #         return {"data_type": DataType(prop_value[0]), "metadata": prop_value[1]}

    def render(self, value):
        if value is None:
            return ""
        eq = " = " if self.eq else " "
        return f"{self.label}{eq}{str(value)}"


class EnumProp(Prop):
    def __init__(self, label, enum_or_list, quoted=False, **kwargs):
        self.enum_type = type(enum_or_list[0]) if isinstance(enum_or_list, list) else enum_or_list
        self.valid_values = set(enum_or_list)
        self.quoted = quoted
        value_expr = pp.MatchFirst([Keywords(val.value) for val in self.valid_values]) | (~Keyword("NULL") + ANY())
        super().__init__(label, value_expr, **kwargs)

    def typecheck(self, prop_value):
        if isinstance(prop_value, list):
            prop_value = prop_value[0]
        prop_value = self.enum_type(prop_value)
        if prop_value not in self.valid_values:
            raise ValueError(f"Invalid value: {prop_value} must be one of {self.valid_values}")
        return prop_value

    def render(self, value):
        if value is None:
            return ""
        eq = " = " if self.eq else " "
        if self.quoted:
            value = f"'{value}'"
        return f"{self.label}{eq}{value}"


class EnumListProp(Prop):
    def __init__(self, label, enum_or_list, **kwargs):
        self.enum_type = type(enum_or_list[0]) if isinstance(enum_or_list, list) else enum_or_list
        self.valid_values = set(enum_or_list)
        enum_values = pp.MatchFirst([Keywords(val.value) for val in self.valid_values])
        value_expr = pp.delimited_list(enum_values | ANY())
        super().__init__(label, value_expr, **kwargs)

    def typecheck(self, prop_values):
        prop_values = [self.enum_type(val) for val in prop_values]
        for value in prop_values:
            if value not in self.valid_values:
                raise ValueError(f"Invalid value: {value} must be one of {self.valid_values}")
        return prop_values

    def render(self, values):
        if values is None or len(values) == 0:
            return ""
        eq = " = " if self.eq else " "
        value_list = ", ".join([str(val) for val in values])
        return f"{self.label}{eq}{value_list}"


class EnumFlagProp(Prop):
    def __init__(self, enum_or_list, **kwargs):
        self.enum_type = type(enum_or_list[0]) if isinstance(enum_or_list, list) else enum_or_list
        self.valid_values = set(enum_or_list)
        value_expr = pp.MatchFirst([Keywords(val.value) for val in self.valid_values])
        super().__init__(label=None, value_expr=value_expr, eq=False, **kwargs)

    def typecheck(self, prop_value):
        prop_value = self.enum_type(prop_value)
        if prop_value not in self.valid_values:
            raise ValueError(f"Invalid value: {prop_value} must be one of {self.valid_values}")
        return prop_value

    def render(self, value):
        if value is None:
            return ""
        return value


class QueryProp(Prop):
    def __init__(self, label):
        value_expr = pp.Word(pp.printables + " \n")
        super().__init__(label, value_expr, eq=False)

    def typecheck(self, prop_value):
        return prop_value

    def render(self, value):
        if value is None:
            return ""
        return f"{self.label} {value}"


class ExpressionProp(Prop):
    def __init__(self, label):
        value_expr = pp.Empty() + pp.SkipTo(Keyword("AS"))("prop_value")
        super().__init__(label, value_expr, eq=False)

    def typecheck(self, prop_value):
        return prop_value.strip()

    def render(self, value):
        if value is None:
            return ""
        return f"{self.label} {value}"


class TimeTravelProp(Prop):
    """
    { AT | BEFORE } ( {
        TIMESTAMP => <timestamp> |
        OFFSET => <time_difference> |
        STATEMENT => <id> |
        STREAM => '<name>'
    } )
    """

    def __init__(self, label):
        value_expr = ANY() + ARROW + ANY()
        super().__init__(label, value_expr, eq=False, parens=True)

    def typecheck(self, prop_value):
        key, value = prop_value
        return {key: value}

    def render(self, values):
        if values is None:
            return ""

        key, value = values.popitem()
        if key.upper() == "STREAM":
            value = f"'{value}'"
        time_point = f"{key} => {value}"
        return f"{self.label} ({time_point})"


class AlertConditionProp(Prop):
    def __init__(self):
        label = "IF"
        value_expr = Keyword("EXISTS").suppress() + pp.original_text_for(pp.nested_expr())("prop_value")
        super().__init__(label, value_expr, eq=False, parens=True)

    def typecheck(self, prop_value):
        return prop_value.strip("()").strip()

    def render(self, value):
        if value is None:
            return ""
        return f"IF(EXISTS( {value} ))"


class SessionParametersProp(Prop):
    pass


class ArgsProp(Prop):
    def __init__(self):
        value_expr = pp.original_text_for(pp.nested_expr())
        super().__init__(label=None, value_expr=value_expr, eq=False)

    def typecheck(self, prop_values):
        arg_parser = pp.delimited_list(
            pp.Group(
                (Identifier | pp.dbl_quoted_string)("name")
                + ANY("data_type")
                + pp.Opt(_in_parens(ANY()))("data_type_size")
            )
        )
        prop_values = prop_values.strip("()".strip())
        if prop_values == "":
            return []
        parsed = arg_parser.parse_string(prop_values)
        args = []
        for arg_data in parsed:
            arg = arg_data.as_dict()
            arg["name"] = arg["name"].strip('"')
            arg["data_type"] = DataType(arg["data_type"])
            args.append(arg)
        return args

    def render(self, value):
        if value is None or len(value) == 0:
            return "()"
        args = []
        for arg in value:
            default = f" DEFAULT {arg['default']}" if arg.get("default") else ""
            args.append(f"{arg['name']} {str(arg['data_type'])}{default}")
        return f"({', '.join(args)})"


class ColumnNamesProp(Prop):
    def __init__(self):
        value_expr = pp.original_text_for(pp.nested_expr())
        super().__init__(label=None, value_expr=value_expr, eq=False)

    def typecheck(self, prop_values):
        prop_values = prop_values.strip("()")
        column_name_parser = pp.delimited_list(
            pp.Group(
                (Identifier() | pp.dbl_quoted_string)("name")
                + pp.Opt(Keyword("COMMENT") + pp.sgl_quoted_string("comment"))
            )
        )
        parsed = column_name_parser.parse_string(prop_values)
        columns = []
        for column_data in parsed:
            column = column_data.as_dict()
            column["name"] = column["name"].strip('"')
            if "comment" in column:
                column["comment"] = column["comment"].strip("'")
            columns.append(column)
        return columns

    def render(self, values):
        if values is None or len(values) == 0:
            return "()"
        columns = []
        for column in values:
            name = column["name"]
            comment = f" COMMENT '{column['comment']}'" if "comment" in column else ""
            column_str = f"{name}{comment}"
            columns.append(column_str)
        return f"({', '.join(columns)})"


class SchemaProp(Prop):
    def __init__(self):
        super().__init__(label=None, value_expr=pp.NoMatch())

    def typecheck(self, prop_values):
        pass

    def render(self, values):
        if values is None or len(values) == 0:
            return "()"
        columns = []
        for column in values:
            name = column["name"]
            data_type = str(column["data_type"])
            not_null = " NOT NULL" if column["not_null"] else ""

            if isinstance(column["default"], str):
                default = f" DEFAULT '{column['default']}'"
            elif column["default"] is not None:
                default = f" DEFAULT {column['default']}"
            else:
                default = ""

            comment = f" COMMENT '{column['comment']}'" if "comment" in column and column["comment"] else ""
            column_str = f"{name} {data_type}{not_null}{default}{comment}"
            columns.append(column_str)
        return f"({', '.join(columns)})"



from inflection import pluralize

from .enums import ResourceType
from .identifiers import resource_label_for_type
from .resource_name import ResourceName
from .resources.resource import ResourcePointer
from .resources import (
    Database,
    Grant,
    RoleGrant,
    Schema,
    Resource,
    User,
)


def resources_from_role_grants_config(role_grants_config: list) -> list:
    if len(role_grants_config) == 0:
        return []
    resources = []
    for role_grant in role_grants_config:

        if "to_role" in role_grant:
            resources.append(
                RoleGrant(
                    role=role_grant["role"],
                    to_role=role_grant["to_role"],
                )
            )
        elif "to_user" in role_grant:
            resources.append(
                RoleGrant(
                    role=role_grant["role"],
                    to_user=role_grant["to_user"],
                )
            )
        else:
            for user in role_grant.get("users", []):
                resources.append(
                    RoleGrant(
                        role=role_grant["role"],
                        to_user=user,
                    )
                )
            for to_role in role_grant.get("roles", []):
                resources.append(
                    RoleGrant(
                        role=role_grant["role"],
                        to_role=to_role,
                    )
                )
    if len(resources) == 0:
        raise ValueError(f"No role grants found in config: {role_grants_config}")
    return resources


def resources_from_database_config(databases_config: list) -> list:
    resources = []
    for database in databases_config:
        schemas = database.pop("schemas", [])
        db = Database(**database)
        resources.append(db)
        for schema in schemas:
            sch = Schema(**schema)
            db.add(sch)
            resources.append(sch)
    return resources


def resources_from_grants_config(grants_config: list) -> list:
    resources = []
    for grant in grants_config:
        if isinstance(grant, dict):
            titan_grant = Grant(**grant)
        elif isinstance(grant, str):
            titan_grant = Grant.from_sql(grant)
        else:
            raise Exception(f"Unsupported grant found: {type(grant)}, {grant}")
        resources.append(titan_grant)
    return resources


def resources_from_users_config(users_config: list) -> list:
    resources = []
    for user in users_config:
        if isinstance(user, dict):
            roles = user.pop("roles", [])
            titan_user = User(**user)
            resources.append(titan_user)
            for role in roles:
                resources.append(
                    RoleGrant(
                        role=role,
                        to_user=titan_user,
                    )
                )
        elif isinstance(user, str):
            resources.append(User.from_sql(user))
    return resources


def process_requires(resource: Resource, requires: list):
    for req in requires:
        resource.requires(ResourcePointer(name=req["name"], resource_type=ResourceType(req["resource_type"])))


def collect_resources_from_config(config: dict):
    # TODO: ResourcePointers should get resolved to top-level resource configs when possible

    config = config.copy()

    database_config = config.pop("databases", [])
    role_grants = config.pop("role_grants", [])
    grants = config.pop("grants", [])
    users = config.pop("users", [])

    resources = []

    for resource_type in Resource.__types__.keys():
        resource_label = pluralize(resource_label_for_type(resource_type))
        for data in config.pop(resource_label, []):
            try:
                requires = data.pop("requires", [])
                resource_cls = Resource.resolve_resource_cls(resource_type, data)
                resource = resource_cls(**data)
                process_requires(resource, requires)
                resources.append(resource)
            except Exception as e:
                print(f"Error processing resource: {data}")
                raise e

    if config:
        raise ValueError(f"Unknown keys in config: {config.keys()}")

    resources.extend(resources_from_database_config(database_config))
    resources.extend(resources_from_role_grants_config(role_grants))
    resources.extend(resources_from_grants_config(grants))
    resources.extend(resources_from_users_config(users))

    resource_cache = {}
    for resource in resources:
        if hasattr(resource._data, "name"):
            resource_cache[(resource.resource_type, ResourceName(resource._data.name))] = resource

    for resource in resources:
        if resource.resource_type == ResourceType.GRANT:
            cache_pointer = (resource.on_type, ResourceName(resource.on))
            if cache_pointer in resource_cache:
                resource._data.on = ResourceName(str(resource_cache[cache_pointer].fqn))

        # TODO: investigate this
        # for ref in resource.refs:
        #     cache_pointer = (ref.resource_type, ResourceName(ref.name))
        #     if (
        #         isinstance(ref, ResourcePointer)
        #         and cache_pointer in resource_cache
        #         and resource_cache[cache_pointer]._container is not None
        #     ):
        #         ref._container = resource_cache[cache_pointer]._container

    return resources



import pytest
from inflection import pluralize

from tests.helpers import get_json_fixtures
from titan.gitops import collect_resources_from_config
from titan.identifiers import resource_label_for_type

JSON_FIXTURES = list(get_json_fixtures())


@pytest.fixture
def database_config() -> dict:
    return {
        "databases": [
            {
                "name": "test_database",
                "comment": "test database",
                "schemas": [
                    {
                        "name": "test_schema",
                        "comment": "test schema",
                    }
                ],
            }
        ]
    }


@pytest.fixture
def resource_config() -> dict:
    config = {}
    for resource_cls, resource_config in JSON_FIXTURES:
        config[pluralize(resource_label_for_type(resource_cls.resource_type))] = [resource_config]

    return config


def test_database_config(database_config):
    resources = collect_resources_from_config(database_config)
    assert len(resources) == 2


@pytest.mark.skip(reason="JSON_FIXTURES doesn't include things like role grants yet")
def test_resource_config(resource_config):
    resources = collect_resources_from_config(resource_config)
    resource_types = set([resource.resource_type for resource in resources])
    expected_resource_types = set([resource_cls.resource_type for resource_cls, _ in JSON_FIXTURES])
    assert resource_types == expected_resource_types



from dataclasses import dataclass

from .resource import Resource, ResourceSpec, NamedResource
from .role import Role
from ..enums import ParseableEnum, ResourceType
from ..scope import AccountScope
from ..props import Props, StringProp, BoolProp, EnumProp, StringListProp
from ..resource_name import ResourceName


class StorageProvider(ParseableEnum):
    S3 = "S3"
    AZURE = "AZURE"
    GCS = "GCS"


@dataclass(unsafe_hash=True)
class _S3StorageIntegration(ResourceSpec):
    name: ResourceName
    enabled: bool
    storage_aws_role_arn: str
    storage_allowed_locations: list[str]
    storage_blocked_locations: list[str] = None
    storage_provider: StorageProvider = StorageProvider.S3
    storage_aws_object_acl: str = None
    type: str = "EXTERNAL_STAGE"
    owner: Role = "ACCOUNTADMIN"
    comment: str = None

    def __post_init__(self):
        super().__post_init__()
        if self.type != "EXTERNAL_STAGE":
            raise ValueError("Type must be 'EXTERNAL_STAGE' for S3StorageIntegration")


class S3StorageIntegration(NamedResource, Resource):
    """
    Description:
        Manages the integration of Snowflake with S3 storage.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-storage-integration

    Fields:
        name (string, required): The name of the storage integration.
        enabled (bool, required): Whether the storage integration is enabled. Defaults to True.
        storage_aws_role_arn (string, required): The AWS IAM role ARN to access the S3 bucket.
        storage_allowed_locations (list, required): A list of allowed locations for storage in the format 's3://<bucket>/<path>/'.
        storage_blocked_locations (list): A list of blocked locations for storage in the format 's3://<bucket>/<path>/'. Defaults to an empty list.
        storage_aws_object_acl (string): The ACL policy for objects stored in S3. Defaults to 'bucket-owner-full-control'.
        type (string): The type of storage integration. Defaults to 'EXTERNAL_STAGE'.
        owner (string or Role): The owner role of the storage integration. Defaults to 'ACCOUNTADMIN'.
        comment (string): An optional comment about the storage integration.

    Python:

        ```python
        s3_storage_integration = S3StorageIntegration(
            name="some_s3_storage_integration",
            enabled=True,
            storage_aws_role_arn="arn:aws:iam::123456789012:role/MyS3AccessRole",
            storage_allowed_locations=["s3://mybucket/myfolder/"],
            storage_blocked_locations=["s3://mybucket/myblockedfolder/"],
            storage_aws_object_acl="bucket-owner-full-control",
            comment="This is a sample S3 storage integration."
        )
        ```

    Yaml:

        ```yaml
        s3_storage_integrations:
          - name: some_s3_storage_integration
            enabled: true
            storage_aws_role_arn: "arn:aws:iam::123456789012:role/MyS3AccessRole"
            storage_allowed_locations:
              - "s3://mybucket/myfolder/"
            storage_blocked_locations:
              - "s3://mybucket/myblockedfolder/"
            storage_aws_object_acl: "bucket-owner-full-control"
            comment: "This is a sample S3 storage integration."
        ```
    """

    resource_type = ResourceType.STORAGE_INTEGRATION
    props = Props(
        type=StringProp("type"),
        storage_provider=EnumProp("storage_provider", [StorageProvider.S3]),
        storage_aws_role_arn=StringProp("storage_aws_role_arn"),
        storage_aws_object_acl=StringProp("storage_aws_object_acl"),
        enabled=BoolProp("enabled"),
        storage_allowed_locations=StringListProp("storage_allowed_locations", parens=True),
        storage_blocked_locations=StringListProp("storage_blocked_locations", parens=True),
        comment=StringProp("comment"),
    )
    scope = AccountScope()
    spec = _S3StorageIntegration

    def __init__(
        self,
        name: str,
        enabled: bool,
        storage_aws_role_arn: str,
        storage_allowed_locations: list[str],
        storage_blocked_locations: list[str] = None,
        storage_aws_object_acl: str = None,
        owner: str = "ACCOUNTADMIN",
        comment: str = None,
        **kwargs,
    ):
        kwargs.pop("storage_provider", None)
        kwargs.pop("type", None)
        super().__init__(name, **kwargs)
        self._data: _S3StorageIntegration = _S3StorageIntegration(
            name=self._name,
            enabled=enabled,
            storage_aws_role_arn=storage_aws_role_arn,
            storage_allowed_locations=storage_allowed_locations,
            storage_blocked_locations=storage_blocked_locations,
            storage_aws_object_acl=storage_aws_object_acl,
            owner=owner,
            comment=comment,
        )


@dataclass(unsafe_hash=True)
class _GCSStorageIntegration(ResourceSpec):
    name: ResourceName
    enabled: bool
    storage_allowed_locations: list[str]
    storage_blocked_locations: list[str] = None
    storage_provider: StorageProvider = StorageProvider.GCS
    type: str = "EXTERNAL_STAGE"
    owner: Role = "ACCOUNTADMIN"
    comment: str = None

    def __post_init__(self):
        super().__post_init__()
        if self.type != "EXTERNAL_STAGE":
            raise ValueError("Type must be 'EXTERNAL_STAGE' for GCSStorageIntegration")


class GCSStorageIntegration(NamedResource, Resource):
    """
    Description:
        Manages the integration of Google Cloud Storage (GCS) as an external stage for storing data.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-storage-integration

    Fields:
        name (string, required): The name of the storage integration.
        enabled (bool, required): Specifies whether the storage integration is enabled.
        storage_allowed_locations (list): A list of allowed GCS locations for data storage.
        storage_blocked_locations (list): A list of blocked GCS locations for data storage.
        owner (string or Role): The owner role of the storage integration. Defaults to 'ACCOUNTADMIN'.
        comment (string): An optional comment about the storage integration.

    Python:

        ```python
        gcs_storage_integration = GCSStorageIntegration(
            name="some_gcs_storage_integration",
            enabled=True,
            storage_allowed_locations=['gcs://bucket/path/'],
            storage_blocked_locations=['gcs://bucket/blocked_path/']
        )
        ```

    Yaml:

        ```yaml
        gcs_storage_integrations:
          - name: some_gcs_storage_integration
            enabled: true
            storage_allowed_locations:
              - 'gcs://bucket/path/'
            storage_blocked_locations:
              - 'gcs://bucket/blocked_path/'
        ```
    """

    resource_type = ResourceType.STORAGE_INTEGRATION
    props = Props(
        type=StringProp("type"),
        storage_provider=EnumProp("storage_provider", [StorageProvider.GCS]),
        enabled=BoolProp("enabled"),
        storage_allowed_locations=StringListProp("storage_allowed_locations", parens=True),
        storage_blocked_locations=StringListProp("storage_blocked_locations", parens=True),
        comment=StringProp("comment"),
    )
    scope = AccountScope()
    spec = _GCSStorageIntegration

    def __init__(
        self,
        name: str,
        enabled: bool,
        storage_allowed_locations: list[str],
        storage_blocked_locations: list[str] = None,
        owner: str = "ACCOUNTADMIN",
        comment: str = None,
        **kwargs,
    ):
        kwargs.pop("storage_provider", None)
        kwargs.pop("type", None)
        super().__init__(name, **kwargs)
        self._data: _GCSStorageIntegration = _GCSStorageIntegration(
            name=self._name,
            enabled=enabled,
            storage_allowed_locations=storage_allowed_locations,
            storage_blocked_locations=storage_blocked_locations,
            owner=owner,
            comment=comment,
        )


@dataclass(unsafe_hash=True)
class _AzureStorageIntegration(ResourceSpec):
    name: ResourceName
    enabled: bool
    azure_tenant_id: str
    storage_allowed_locations: list[str]
    storage_blocked_locations: list[str] = None
    storage_provider: StorageProvider = StorageProvider.AZURE
    type: str = "EXTERNAL_STAGE"
    owner: Role = "ACCOUNTADMIN"
    comment: str = None

    def __post_init__(self):
        super().__post_init__()
        if self.type != "EXTERNAL_STAGE":
            raise ValueError("Type must be 'EXTERNAL_STAGE' for _AzureStorageIntegration")


class AzureStorageIntegration(NamedResource, Resource):
    """
    Description:
        Represents an Azure storage integration in Snowflake, which allows Snowflake to access external cloud storage using Azure credentials.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-storage-integration

    Fields:
        name (string, required): The name of the storage integration.
        enabled (bool, required): Specifies whether the storage integration is enabled.
        azure_tenant_id (string, required): The Azure tenant ID associated with the storage integration.
        storage_allowed_locations (list): The cloud storage locations that are allowed.
        storage_blocked_locations (list): The cloud storage locations that are blocked.
        owner (string or Role): The owner role of the storage integration. Defaults to "ACCOUNTADMIN".
        comment (string): A comment about the storage integration.

    Python:

        ```python
        azure_storage_integration = AzureStorageIntegration(
            name="some_azure_storage_integration",
            enabled=True,
            azure_tenant_id="some_tenant_id",
            storage_allowed_locations=["azure://somebucket/somepath/"],
            storage_blocked_locations=["azure://someotherbucket/somepath/"],
            comment="This is an Azure storage integration."
        )
        ```

    Yaml:

        ```yaml
        azure_storage_integrations:
          - name: some_azure_storage_integration
            enabled: true
            azure_tenant_id: some_tenant_id
            storage_allowed_locations:
              - azure://somebucket/somepath/
            storage_blocked_locations:
              - azure://someotherbucket/somepath/
            comment: This is an Azure storage integration.
        ```
    """

    resource_type = ResourceType.STORAGE_INTEGRATION
    props = Props(
        type=StringProp("type"),
        storage_provider=EnumProp("storage_provider", [StorageProvider.AZURE]),
        azure_tenant_id=StringProp("azure_tenant_id"),
        enabled=BoolProp("enabled"),
        storage_allowed_locations=StringListProp("storage_allowed_locations", parens=True),
        storage_blocked_locations=StringListProp("storage_blocked_locations", parens=True),
        comment=StringProp("comment"),
    )
    scope = AccountScope()
    spec = _AzureStorageIntegration

    def __init__(
        self,
        name: str,
        enabled: bool,
        azure_tenant_id: str,
        storage_allowed_locations: list[str],
        storage_blocked_locations: list[str] = None,
        owner: str = "ACCOUNTADMIN",
        comment: str = None,
        **kwargs,
    ):
        kwargs.pop("storage_provider", None)
        kwargs.pop("type", None)
        super().__init__(name, **kwargs)
        self._data: _AzureStorageIntegration = _AzureStorageIntegration(
            name=self._name,
            enabled=enabled,
            azure_tenant_id=azure_tenant_id,
            storage_allowed_locations=storage_allowed_locations,
            storage_blocked_locations=storage_blocked_locations,
            owner=owner,
            comment=comment,
        )


StorageIntegrationMap = {
    StorageProvider.S3: S3StorageIntegration,
    StorageProvider.GCS: GCSStorageIntegration,
    StorageProvider.AZURE: AzureStorageIntegration,
}


def _resolver(data: dict):
    return StorageIntegrationMap[StorageProvider(data["storage_provider"])]


Resource.__resolvers__[ResourceType.STORAGE_INTEGRATION] = _resolver



from dataclasses import dataclass

from ..enums import ParseableEnum, ResourceType
from ..props import (
    ColumnNamesProp,
    EnumProp,
    IdentifierProp,
    Props,
    QueryProp,
    StringProp,
    TagsProp,
)
from ..resource_name import ResourceName
from ..scope import SchemaScope, TableScope
from .resource import NamedResource, Resource, ResourceSpec
from .role import Role
from .tag import TaggableResource
from .warehouse import Warehouse


class RefreshMode(ParseableEnum):
    AUTO = "AUTO"
    FULL = "FULL"
    INCREMENTAL = "INCREMENTAL"


class InitializeBehavior(ParseableEnum):
    ON_CREATE = "ON_CREATE"
    ON_SCHEDULE = "ON_SCHEDULE"


@dataclass(unsafe_hash=True)
class _DynamicTableColumn(ResourceSpec):
    name: str
    comment: str = None


class DynamicTableColumn(Resource):
    resource_type = ResourceType.COLUMN
    props = Props(
        comment=StringProp("comment", eq=False),
    )
    scope = TableScope()
    spec = _DynamicTableColumn
    serialize_inline = True

    def __init__(
        self,
        name: str,
        comment: str = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._data: _DynamicTableColumn = _DynamicTableColumn(
            name,
            comment=comment,
        )


@dataclass(unsafe_hash=True)
class _DynamicTable(ResourceSpec):
    name: ResourceName
    columns: list[DynamicTableColumn]
    target_lag: str
    warehouse: Warehouse
    as_: str
    refresh_mode: RefreshMode = RefreshMode.AUTO
    initialize: InitializeBehavior = InitializeBehavior.ON_CREATE
    comment: str = None
    owner: Role = "SYSADMIN"


class DynamicTable(NamedResource, TaggableResource, Resource):
    """
    Description:
        Represents a dynamic table in Snowflake, which can be configured to refresh automatically,
        fully, or incrementally, and initialized on creation or on a schedule.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-dynamic-table

    Fields:
        name (string, required): The name of the dynamic table.
        columns (list, required): A list of dicts defining the structure of the table.
        target_lag (string): The acceptable lag (delay) for data in the table. Defaults to "DOWNSTREAM".
        warehouse (string or Warehouse, required): The warehouse where the table is stored.
        as_ (string, required): The query used to populate the table.
        refresh_mode (string or RefreshMode): The mode of refreshing the table (AUTO, FULL, INCREMENTAL).
        initialize (string or InitializeBehavior): The behavior when the table is initialized (ON_CREATE, ON_SCHEDULE).
        comment (string): An optional comment for the table.
        owner (string or Role): The owner of the table. Defaults to "SYSADMIN".

    Python:

        ```python
        dynamic_table = DynamicTable(
            name="some_dynamic_table",
            columns=[{"name": "id"}, {"name": "data"}],
            target_lag="1 HOUR",
            warehouse="some_warehouse",
            refresh_mode="AUTO",
            initialize="ON_CREATE",
            as_="SELECT id, data FROM source_table",
            comment="This is a sample dynamic table",
            owner="SYSADMIN"
        )
        ```

    Yaml:

        ```yaml
        dynamic_table:
          name: some_dynamic_table
          columns:
            - name: id
            - name: data
          target_lag: "1 HOUR"
          warehouse: some_warehouse
          refresh_mode: AUTO
          initialize: ON_CREATE
          as_: "SELECT id, data FROM source_table"
          comment: "This is a sample dynamic table"
          owner: SYSADMIN
        ```
    """

    resource_type = ResourceType.DYNAMIC_TABLE
    props = Props(
        columns=ColumnNamesProp(),
        target_lag=StringProp("target_lag", alt_tokens=["DOWNSTREAM"]),
        warehouse=IdentifierProp("warehouse"),
        refresh_mode=EnumProp("refresh_mode", RefreshMode),
        initialize=EnumProp("initialize", InitializeBehavior),
        comment=StringProp("comment"),
        as_=QueryProp("as"),
        tags=TagsProp(),
    )
    scope = SchemaScope()
    spec = _DynamicTable

    def __init__(
        self,
        name: str,
        columns: list[dict],
        target_lag: str,
        warehouse: str,
        as_: str,
        refresh_mode: RefreshMode = RefreshMode.AUTO,
        initialize: InitializeBehavior = InitializeBehavior.ON_CREATE,
        comment: str = None,
        owner: str = "SYSADMIN",
        tags: dict[str, str] = None,
        **kwargs,
    ):
        super().__init__(name, **kwargs)
        self._data: _DynamicTable = _DynamicTable(
            name=self._name,
            columns=columns,
            target_lag=target_lag,
            warehouse=warehouse,
            refresh_mode=refresh_mode,
            initialize=initialize,
            as_=as_,
            comment=comment,
            owner=owner,
        )
        self.set_tags(tags)



from dataclasses import dataclass

from .resource import Resource, ResourceSpec, NamedResource
from .role import Role
from ..enums import ParseableEnum, ResourceType
from ..scope import AccountScope
from ..resource_name import ResourceName
from ..props import (
    EnumProp,
    IntProp,
    Props,
    StringProp,
    StringListProp,
)


class ResourceMonitorFrequency(ParseableEnum):
    MONTHLY = "MONTHLY"
    DAILY = "DAILY"
    WEEKLY = "WEEKLY"
    YEARLY = "YEARLY"
    NEVER = "NEVER"


@dataclass(unsafe_hash=True)
class _ResourceMonitor(ResourceSpec):
    name: ResourceName
    owner: Role = "ACCOUNTADMIN"
    credit_quota: int = None
    frequency: ResourceMonitorFrequency = None
    start_timestamp: str = None
    end_timestamp: str = None
    notify_users: list[str] = None

    def __post_init__(self):
        super().__post_init__()
        if self.credit_quota is not None and not isinstance(self.credit_quota, int):
            raise ValueError("credit_quota must be an integer or None")
        if self.start_timestamp and self.frequency is None:
            self.frequency = ResourceMonitorFrequency.MONTHLY
        if self.owner.name != "ACCOUNTADMIN":
            raise ValueError("ResourceMonitors can only be created by ACCOUNTADMIN")


class ResourceMonitor(NamedResource, Resource):
    """
    Description:
        Manages the monitoring of resource usage within an account.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-resource-monitor

    Fields:
        name (string, required): The name of the resource monitor.
        credit_quota (int): The amount of credits that can be used by this monitor. Defaults to None.
        frequency (string or ResourceMonitorFrequency): The frequency of monitoring. Defaults to None.
        start_timestamp (string): The start time for the monitoring period. Defaults to None.
        end_timestamp (string): The end time for the monitoring period. Defaults to None.
        notify_users (list): A list of users to notify when thresholds are reached. Defaults to None.

    Python:

        ```python
        resource_monitor = ResourceMonitor(
            name="some_resource_monitor",
            credit_quota=1000,
            frequency="DAILY",
            start_timestamp="2049-01-01 00:00",
            end_timestamp="2049-12-31 23:59",
            notify_users=["user1", "user2"]
        )
        ```

    Yaml:

        ```yaml
        resource_monitors:
          - name: some_resource_monitor
            credit_quota: 1000
            frequency: DAILY
            start_timestamp: "2049-01-01 00:00"
            end_timestamp: "2049-12-31 23:59"
            notify_users:
              - user1
              - user2
        ```
    """

    resource_type = ResourceType.RESOURCE_MONITOR
    props = Props(
        _start_token="WITH",
        credit_quota=IntProp("credit_quota"),
        frequency=EnumProp("frequency", ResourceMonitorFrequency),
        start_timestamp=StringProp("start_timestamp", alt_tokens=["IMMEDIATELY"]),
        end_timestamp=StringProp("end_timestamp"),
        notify_users=StringListProp("notify_users", parens=True),
    )
    scope = AccountScope()
    spec = _ResourceMonitor

    def __init__(
        self,
        name: str,
        credit_quota: int = None,
        frequency: ResourceMonitorFrequency = None,
        start_timestamp: str = None,
        end_timestamp: str = None,
        notify_users: list[str] = None,
        owner: str = "ACCOUNTADMIN",
        **kwargs,
    ):
        super().__init__(name, **kwargs)
        self._data: _ResourceMonitor = _ResourceMonitor(
            name=self._name,
            credit_quota=credit_quota,
            frequency=frequency,
            start_timestamp=start_timestamp,
            end_timestamp=end_timestamp,
            notify_users=notify_users,
            owner=owner,
        )
        # TODO: rely on notify_users



from dataclasses import dataclass

from .resource import Resource, ResourceSpec, NamedResource
from .role import Role
from ..enums import ParseableEnum, ResourceType
from ..scope import AccountScope
from ..resource_name import ResourceName
from ..props import (
    EnumListProp,
    FlagProp,
    IdentifierListProp,
    Props,
    StringProp,
)


class ObjectType(ParseableEnum):
    ACCOUNT_PARAMETERS = "ACCOUNT PARAMETERS"
    DATABASES = "DATABASES"
    INTEGRATIONS = "INTEGRATIONS"
    NETWORK_POLICIES = "NETWORK POLICIES"
    RESOURCE_MONITORS = "RESOURCE MONITORS"
    ROLES = "ROLES"
    SHARES = "SHARES"
    USERS = "USERS"
    WAREHOUSES = "WAREHOUSES"


class IntegrationTypes(ParseableEnum):
    SECURITY_INTEGRATIONS = "SECURITY INTEGRATIONS"
    API_INTEGRATIONS = "API INTEGRATIONS"
    NOTIFICATION_INTEGRATIONS = "NOTIFICATION INTEGRATIONS"


@dataclass(unsafe_hash=True)
class _FailoverGroup(ResourceSpec):
    name: ResourceName
    object_types: list[ObjectType]
    allowed_accounts: list[str]
    allowed_databases: list[str] = None
    allowed_shares: list[str] = None
    allowed_integration_types: list[IntegrationTypes] = None
    ignore_edition_check: bool = None
    replication_schedule: str = None
    owner: Role = "ACCOUNTADMIN"


class FailoverGroup(NamedResource, Resource):
    """
    Description:
        Represents a failover group in Snowflake, which is a collection of databases, shares, and other resources
        that can be failed over together to a secondary account in case of a disaster recovery scenario.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-failover-group

    Fields:
        name (string, required): The name of the failover group.
        object_types (list): The types of objects included in the failover group. Can include string or ObjectType.
        allowed_accounts (list, required): The accounts that are allowed to be part of the failover group.
        allowed_databases (list): The databases that are allowed to be part of the failover group.
        allowed_shares (list): The shares that are allowed to be part of the failover group.
        allowed_integration_types (list): The integration types that are allowed in the failover group. Can include string or IntegrationTypes.
        ignore_edition_check (bool): Specifies whether to ignore the edition check. Defaults to None.
        replication_schedule (string): The schedule for replication. Defaults to None.
        owner (string or Role): The owner role of the failover group. Defaults to "ACCOUNTADMIN".

    Python:

        ```python
        failover_group = FailoverGroup(
            name="some_failover_group",
            object_types=["DATABASES", "ROLES"],
            allowed_accounts=["org1.account1", "org2.account2"],
            allowed_databases=["db1", "db2"],
            allowed_shares=["share1", "share2"],
            allowed_integration_types=["SECURITY INTEGRATIONS", "API INTEGRATIONS"],
            ignore_edition_check=True,
            replication_schedule="USING CRON 0 0 * * * UTC",
            owner="ACCOUNTADMIN"
        )
        ```

    Yaml:

        ```yaml
        failover_groups:
          - name: some_failover_group
            object_types:
              - DATABASES
              - ROLES
            allowed_accounts:
              - org1.account1
              - org2.account2
            allowed_databases:
              - db1
              - db2
            allowed_shares:
              - share1
              - share2
            allowed_integration_types:
              - SECURITY INTEGRATIONS
              - API INTEGRATIONS
            ignore_edition_check: true
            replication_schedule: "USING CRON 0 0 * * * UTC"
            owner: ACCOUNTADMIN
        ```
    """

    resource_type = ResourceType.FAILOVER_GROUP
    props = Props(
        object_types=EnumListProp("object_types", ObjectType),
        allowed_databases=IdentifierListProp("allowed_databases"),
        allowed_shares=IdentifierListProp("allowed_shares"),
        allowed_integration_types=EnumListProp("allowed_integration_types", IntegrationTypes),
        allowed_accounts=IdentifierListProp("allowed_accounts"),
        ignore_edition_check=FlagProp("ignore edition check"),
        replication_schedule=StringProp("replication_schedule"),
    )
    scope = AccountScope()
    spec = _FailoverGroup

    def __init__(
        self,
        name: str,
        object_types: list[ObjectType],
        allowed_accounts: list[str],
        allowed_databases: list[str] = None,
        allowed_shares: list[str] = None,
        allowed_integration_types: list[IntegrationTypes] = None,
        ignore_edition_check: bool = None,
        replication_schedule: str = None,
        owner: str = "ACCOUNTADMIN",
        **kwargs,
    ):
        super().__init__(name, **kwargs)
        self._data: _FailoverGroup = _FailoverGroup(
            name=self._name,
            object_types=object_types,
            allowed_accounts=allowed_accounts,
            allowed_databases=allowed_databases,
            allowed_shares=allowed_shares,
            allowed_integration_types=allowed_integration_types,
            ignore_edition_check=ignore_edition_check,
            replication_schedule=replication_schedule,
            owner=owner,
        )



import re
from functools import lru_cache
from typing import Union

import pyparsing as pp
import yaml

from .parse_primitives import FullyQualifiedIdentifier


@lru_cache(maxsize=1024 * 1024)
def attribute_is_resource_name(attribute: str) -> bool:
    return attribute == "name" or attribute == "on" or attribute.endswith("_name") or attribute == "owner"


@lru_cache(maxsize=1024 * 1024)
def resource_name_from_snowflake_metadata(name: str) -> "ResourceName":
    if name.startswith('"') and name.endswith('"'):
        raise RuntimeError(f"{name} is not from snowflake metadata")
    if re.match(r"^[A-Z_][A-Z0-9_]*$", name):
        return ResourceName(name)
    else:
        return ResourceName(f'"{name}"')


@lru_cache(maxsize=1024 * 1024)
def _name_should_be_quoted(name: str) -> bool:
    try:
        # If we can parse it, we don't need to quote it
        FullyQualifiedIdentifier.parse_string(name, parse_all=True)
        return False
    except pp.ParseException:
        return True


class ResourceName:
    def __init__(self, name: Union[str, "ResourceName"]) -> None:
        if not isinstance(name, (str, ResourceName)):
            raise ValueError(f"ResourceName must be a string or ResourceName, got {type(name)}")

        if isinstance(name, ResourceName):
            self._name = name._name
            self._quoted = name._quoted
        elif name.startswith('"') and name.endswith('"'):
            self._name = name[1:-1]
            self._quoted = True
        else:
            self._name = name
            self._quoted = _name_should_be_quoted(name)

    def __repr__(self):
        name = getattr(self, "_name", None)
        quoted = getattr(self, "_quoted", False)
        name = f'"{name}"' if quoted else name
        return f"Resource:{name}"

    def __hash__(self):
        return hash(str(self))

    def __str__(self):
        return f'"{self._name}"' if self._quoted else self._name.upper()

    def __eq__(self, other: Union[str, "ResourceName"]):
        if not isinstance(other, (ResourceName, str)):
            return False
        if isinstance(other, str):
            other = ResourceName(other)
        if self._quoted and other._quoted:
            return self._name == other._name
        elif not self._quoted and not other._quoted:
            return self._name.upper() == other._name.upper()
        elif self._quoted and not other._quoted:
            return self._name == other._name.upper()
        elif not self._quoted and other._quoted:
            return self._name.upper() == other._name
        else:
            raise ValueError("Cannot compare ResourceName with a non-ResourceName")

    # @classmethod
    # def from_snowflake_metadata(cls, name: str) -> "ResourceName":
    #     if name.startswith('"') and name.endswith('"'):
    #         raise RuntimeError(f"{name} is not from snowflake metadata")
    #     if re.match(r"^[A-Z_][A-Z0-9_]*$", name):
    #         return cls(name)
    #     else:
    #         return cls(f'"{name}"')

    def upper(self):
        return self

    def startswith(self, prefix: str) -> bool:
        return self._name.startswith(prefix)


yaml.add_representer(ResourceName, lambda dumper, data: dumper.represent_str(str(data)))



from dataclasses import dataclass

from ..enums import ResourceType
from ..props import (
    IdentifierListProp,
    Props,
    StringListProp,
    StringProp,
)
from ..resource_name import ResourceName
from ..scope import AccountScope
from .network_rule import NetworkRule
from .resource import NamedResource, Resource, ResourceSpec
from .role import Role


@dataclass(unsafe_hash=True)
class _NetworkPolicy(ResourceSpec):
    name: ResourceName
    allowed_network_rule_list: list[NetworkRule] = None
    blocked_network_rule_list: list[NetworkRule] = None
    allowed_ip_list: list[str] = None
    blocked_ip_list: list[str] = None
    comment: str = None
    owner: Role = "SECURITYADMIN"

    def __post_init__(self):
        super().__post_init__()
        if self.allowed_network_rule_list is not None and len(self.allowed_network_rule_list) == 0:
            raise ValueError("allowed_network_rule_list must have at least one entry")
        if self.blocked_network_rule_list is not None and len(self.blocked_network_rule_list) == 0:
            raise ValueError("blocked_network_rule_list must have at least one entry")


class NetworkPolicy(NamedResource, Resource):
    """
    Description:
        A Network Policy in Snowflake defines a set of network rules and IP addresses
        that are allowed or blocked from accessing a Snowflake account. This helps in
        managing network traffic and securing access based on network policies.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-network-policy

    Fields:
        name (string, required): The name of the network policy.
        allowed_network_rule_list (list): A list of allowed network rules.
        blocked_network_rule_list (list): A list of blocked network rules.
        allowed_ip_list (list): A list of allowed IP addresses.
        blocked_ip_list (list): A list of blocked IP addresses.
        comment (string): A comment about the network policy.
        owner (string or Role): The owner role of the network policy. Defaults to "SECURITYADMIN".

    Python:

        ```python
        network_policy = NetworkPolicy(
            name="some_network_policy",
            allowed_network_rule_list=[NetworkRule(name="rule1"), NetworkRule(name="rule2")],
            blocked_network_rule_list=[NetworkRule(name="rule3")],
            allowed_ip_list=["192.168.1.1", "192.168.1.2"],
            blocked_ip_list=["10.0.0.1"],
            comment="Example network policy"
        )
        ```

    Yaml:

        ```yaml
        network_policies:
          - name: some_network_policy
            allowed_network_rule_list:
              - rule1
              - rule2
            blocked_network_rule_list:
              - rule3
            allowed_ip_list: ["192.168.1.1", "192.168.1.2"]
            blocked_ip_list: ["10.0.0.1"]
            comment: "Example network policy"
        ```
    """

    resource_type = ResourceType.NETWORK_POLICY
    props = Props(
        allowed_network_rule_list=IdentifierListProp("allowed_network_rule_list", parens=True, eq=True),
        blocked_network_rule_list=IdentifierListProp("blocked_network_rule_list", parens=True, eq=True),
        allowed_ip_list=StringListProp("allowed_ip_list", parens=True, eq=True),
        blocked_ip_list=StringListProp("blocked_ip_list", parens=True, eq=True),
        comment=StringProp("comment"),
    )
    scope = AccountScope()
    spec = _NetworkPolicy

    def __init__(
        self,
        name: str,
        allowed_network_rule_list: list[NetworkRule] = None,
        blocked_network_rule_list: list[NetworkRule] = None,
        allowed_ip_list: list[str] = None,
        blocked_ip_list: list[str] = None,
        comment: str = None,
        owner: str = "SECURITYADMIN",
        **kwargs,
    ):
        super().__init__(name, **kwargs)
        self._data: _NetworkPolicy = _NetworkPolicy(
            name=self._name,
            allowed_network_rule_list=allowed_network_rule_list,
            blocked_network_rule_list=blocked_network_rule_list,
            allowed_ip_list=allowed_ip_list,
            blocked_ip_list=blocked_ip_list,
            comment=comment,
            owner=owner,
        )
        if self._data.allowed_network_rule_list:
            for network_rule in self._data.allowed_network_rule_list:
                self.requires(network_rule)
        if self._data.blocked_network_rule_list:
            for network_rule in self._data.blocked_network_rule_list:
                self.requires(network_rule)



"Adapted from https://github.com/snowflakedb/snowflake-cli/blob/main/src/snowflake/cli/app/snow_connector.py"

# Copyright (c) 2024 Snowflake Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import errno
import logging
import os
import shutil
import tempfile

from contextlib import contextmanager, redirect_stdout, redirect_stderr
from pathlib import Path
from typing import Optional, Union, Dict

import snowflake.connector

from click.exceptions import ClickException
from snowflake.connector import SnowflakeConnection
from snowflake.connector.errors import DatabaseError, ForbiddenError

log = logging.getLogger()

DEFAULT_SIZE_LIMIT_MB = 128
UNLIMITED = -1
ENCRYPTED_PKCS8_PK_HEADER = b"-----BEGIN ENCRYPTED PRIVATE KEY-----"
UNENCRYPTED_PKCS8_PK_HEADER = b"-----BEGIN PRIVATE KEY-----"


class InvalidConnectionConfiguration(ClickException):
    def format_message(self):
        return f"Invalid connection configuration. {self.message}"


class SnowflakeConnectionError(ClickException):
    def __init__(self, snowflake_err: Exception):
        super().__init__(f"Could not connect to Snowflake. Reason: {snowflake_err}")


class DirectoryIsNotEmptyError(ClickException):
    def __init__(self, path: Path):
        super().__init__(f"Directory '{path}' is not empty")


class FileTooLargeError(ClickException):
    def __init__(self, path: Path, size_limit_in_kb: int):
        super().__init__(f"File {path} is too large (size limit: {size_limit_in_kb} KB)")


def connect(
    mfa_passcode: Optional[str] = None,
    enable_diag: Optional[bool] = False,
    diag_log_path: Optional[str] = None,
    diag_allowlist_path: Optional[str] = None,
    **kwargs,
) -> SnowflakeConnection:

    using_session_token = "session_token" in kwargs and kwargs["session_token"] is not None
    using_master_token = "master_token" in kwargs and kwargs["master_token"] is not None
    _raise_errors_related_to_session_token(using_session_token, using_master_token)

    connection_parameters = get_env_vars()

    # Apply kwargs to connection details
    for key, value in kwargs.items():
        # Command line override case
        if value:
            connection_parameters[key] = value
            continue

        # Generic environment variable case, apply only if value not passed via flag or connection variable
        generic_env_value = os.environ.get(f"SNOWFLAKE_{key}".upper())
        if key not in connection_parameters and generic_env_value:
            connection_parameters[key] = generic_env_value
            continue

    # Clean up connection params
    connection_parameters = {k: v for k, v in connection_parameters.items() if v is not None}

    connection_parameters = _update_connection_details_with_private_key(connection_parameters)

    if mfa_passcode:
        connection_parameters["passcode"] = mfa_passcode

    if connection_parameters.get("authenticator") == "username_password_mfa":
        connection_parameters["client_request_mfa_token"] = True

    if enable_diag:
        connection_parameters["enable_connection_diag"] = enable_diag
        if diag_log_path:
            connection_parameters["connection_diag_log_path"] = diag_log_path
        if diag_allowlist_path:
            connection_parameters["connection_diag_allowlist_path"] = diag_allowlist_path

    # Make sure the connection is not closed if it was shared to the SnowCLI, instead of being created in the SnowCLI
    _avoid_closing_the_connection_if_it_was_shared(using_session_token, using_master_token, connection_parameters)

    _update_connection_application_name(connection_parameters)

    try:
        # Whatever output is generated when creating connection,
        # we don't want it in our output. This is particularly important
        # for cases when external browser and json format are used.
        # Redirecting both stdout and stderr for offline usage.
        with redirect_stdout(None), redirect_stderr(None):
            return snowflake.connector.connect(
                **connection_parameters,
            )
    except ForbiddenError as err:
        raise SnowflakeConnectionError(err)
    except DatabaseError as err:
        raise InvalidConnectionConfiguration(err.msg or "")


class SecurePath:
    def __init__(self, path: Union[Path, str]):
        self._path = Path(path)

    def __repr__(self):
        return f'SecurePath("{self._path}")'

    def __truediv__(self, key):
        return SecurePath(self._path / key)

    @property
    def path(self) -> Path:
        """
        Returns itself in pathlib.Path format
        """
        return self._path

    @property
    def parent(self):
        """
        The logical parent of the path. For details, check pathlib.Path.parent
        """
        return SecurePath(self._path.parent)

    def absolute(self):
        """
        Make the path absolute, without normalization or resolving symlinks.
        """
        return SecurePath(self._path.absolute())

    def iterdir(self):
        """
        When the path points to a directory, yield path objects of the directory contents.
        Otherwise, NotADirectoryError is raised.
        If the location does not exist, FileNotFoundError is raised.

        For details, check pathlib.Path.iterdir()
        """
        self.assert_exists()
        self.assert_is_directory()
        return (SecurePath(p) for p in self._path.iterdir())

    def exists(self) -> bool:
        """
        Return True if the path points to an existing file or directory.
        """
        return self._path.exists()

    def is_dir(self) -> bool:
        """
        Return True if the path points to a directory (or a symbolic link pointing to a directory),
        False if it points to another kind of file.
        """
        return self._path.is_dir()

    def is_file(self) -> bool:
        """
        Return True if the path points to a regular file (or a symbolic link pointing to a regular file),
        False if it points to another kind of file.
        """
        return self._path.is_file()

    @property
    def name(self) -> str:
        """A string representing the final path component."""
        return self._path.name

    def chmod(self, permissions_mask: int) -> None:
        """
        Change the file mode and permissions, like os.chmod().
        """
        log.info("Update permissions of file %s to %s", self._path, oct(permissions_mask))
        self._path.chmod(permissions_mask)

    def restrict_permissions(self) -> None:
        """
        Restrict file/directory permissions to owner-only.
        """
        import stat

        owner_permissions = (
            # https://docs.python.org/3/library/stat
            stat.S_IRUSR  # readable by owner
            | stat.S_IWUSR  # writeable by owner
            | stat.S_IXUSR  # executable by owner
        )
        self.chmod(self._path.stat().st_mode & owner_permissions)

    def touch(self, permissions_mask: int = 0o600, exist_ok: bool = True) -> None:
        """
        Create a file at this given path. For details, check pathlib.Path.touch()
        """
        if not self.exists():
            log.info("Creating file %s", str(self._path))
        self._path.touch(mode=permissions_mask, exist_ok=exist_ok)

    def mkdir(
        self,
        permissions_mask: int = 0o700,
        parents: bool = False,
        exist_ok: bool = False,
    ) -> None:
        """
        Create a directory at this given path. For details, check pathlib.Path.mkdir()
        """
        if parents and not self.parent.exists():
            self.parent.mkdir(permissions_mask=permissions_mask, exist_ok=exist_ok, parents=True)
        if not self.exists():
            log.info("Creating directory %s", str(self._path))
        self._path.mkdir(mode=permissions_mask, exist_ok=exist_ok)

    def read_text(self, file_size_limit_mb: int, *args, **kwargs) -> str:
        """
        Return the decoded contents of the file as a string.
        Raises an error of the file exceeds the specified size limit.
        For details, check pathlib.Path.read_text()
        """
        self._assert_exists_and_is_file()
        self._assert_file_size_limit(file_size_limit_mb)
        log.info("Reading file %s", self._path)
        return self._path.read_text(*args, **kwargs)

    def write_text(self, *args, **kwargs):
        """
        Open the file pointed to in text mode, write data to it, and close the file.
        """
        if not self.exists():
            self.touch()
        log.info("Writing to file %s", self._path)
        self.path.write_text(*args, **kwargs)

    @contextmanager
    def open(  # noqa: A003
        self,
        mode="r",
        read_file_limit_mb: Optional[int] = None,
        **open_kwargs,
    ):
        """
        Open the file pointed by this path and return a file object, as
        the built-in open() function does.
        If the file is opened for reading, [read_file_limit_kb] parameter must be provided.
        Raises error if the read file exceeds the specified size limit.
        """
        opened_for_reading = "r" in mode
        if opened_for_reading:
            assert read_file_limit_mb is not None, "For reading mode ('r') read_file_limit_mb must be provided"
            self._assert_exists_and_is_file()
            self._assert_file_size_limit(read_file_limit_mb)

        if self.exists():
            self.assert_is_file()
        else:
            self.touch()  # makes sure permissions of freshly-created file are strict

        log.info("Opening file %s in mode '%s'", self._path, mode)
        with self._path.open(mode=mode, **open_kwargs) as fd:
            yield fd
        log.info("Closing file %s", self._path)

    def move(self, destination: Union[Path, str]) -> "SecurePath":
        """Recursively move a file or directory (src) to another location and return the destination.

        If dst is an existing directory or a symlink to a directory, then src is moved inside that directory.
        The destination path in that directory must not already exist.
        """
        destination = Path(destination)
        if destination.is_dir():
            destination = destination / self._path.name
        if destination.exists():
            _raise_file_exists_error(destination)
        log.info("Moving %s to %s", str(self._path), destination.resolve())
        return SecurePath(shutil.move(str(self._path), destination))

    def copy(self, destination: Union[Path, str], dirs_exist_ok: bool = False) -> "SecurePath":
        """
        Copy the file/directory into the destination.
        If source is a directory, its whole content is copied recursively.
        Permissions of the copy are limited only to the owner.

        If destination is an existing directory, the copy will be created inside it,
        unless dirs_exist_ok is true and the destination has the same name as this path.

        Otherwise, the copied file/base directory will be renamed to match destination.
        If dirs_exist_ok is false (the default) and dst already exists,
        a FileExistsError is raised. If dirs_exist_ok is true,
        the copying operation will continue if it encounters existing directories,
        and files within the destination tree will be overwritten by corresponding
        files from the src tree.
        """
        self.assert_exists()

        destination = Path(destination)
        if destination.exists():
            if destination.is_dir() and (destination.name != self._path.name or self.path.is_file()):
                destination = destination / self._path.name

            if destination.exists():
                if not all([destination.is_dir(), self._path.is_dir(), dirs_exist_ok]):
                    raise FileExistsError(errno.EEXIST, os.strerror(errno.EEXIST), self._path.resolve())

        def _recursive_check_for_conflicts(src: Path, dst: Path):
            if dst.exists() and not dirs_exist_ok:
                _raise_file_exists_error(dst)
            if dst.is_file() and not src.is_file():
                _raise_not_a_directory_error(dst)
            if dst.is_dir() and not src.is_dir():
                _raise_is_a_directory_error(dst)
            if src.is_dir():
                for child in src.iterdir():
                    _recursive_check_for_conflicts(child, dst / child.name)

        def _recursive_copy(src: SecurePath, dst: SecurePath):
            if src.is_file():
                log.info("Copying file %s into %s", src.path, dst.path)
                if dst.exists():
                    dst.unlink()
                shutil.copyfile(src.path, dst.path)
                dst.restrict_permissions()
            if src.is_dir():
                dst.mkdir(exist_ok=True)
                for child in src.iterdir():
                    _recursive_copy(child, dst / child.name)

        _recursive_check_for_conflicts(self._path, destination)
        _recursive_copy(self, self.__class__(destination))

        return SecurePath(destination)

    def unlink(self, missing_ok=False):
        """
        Remove this file or symbolic link.
        If the path points to a directory, use SecurePath.rmdir() instead.

        Check pathlib.Path.unlink() for details.
        """
        if not self.exists():
            if not missing_ok:
                self.assert_exists()
            return

        self.assert_is_file()
        log.info("Removing file %s", self._path)
        self._path.unlink()

    def rmdir(self, recursive=False, missing_ok=False):
        """
        Remove this directory.
        If the path points to a file, use SecurePath.unlink() instead.

        If path points to a file, NotADirectoryError will be raised.
        If directory does not exist, FileNotFoundError will be raised unless [missing_ok] is True.
        If the directory is not empty, DirectoryNotEmpty will be raised unless [recursive] is True.
        """
        if not self.exists():
            if not missing_ok:
                self.assert_exists()
            return

        self.assert_is_directory()

        if not recursive and any(self._path.iterdir()):
            raise DirectoryIsNotEmptyError(self._path.resolve())

        log.info("Removing directory %s", self._path)
        shutil.rmtree(str(self._path))

    @classmethod
    @contextmanager
    def temporary_directory(cls):
        """
        Creates a temporary directory in the most secure manner possible.
        The directory is readable, writable, and searchable only by the creating user ID.
        Yields SecurePath pointing to the absolute location of created directory.

        Works similarly to tempfile.TemporaryDirectory
        """
        with tempfile.TemporaryDirectory(prefix="snowflake-cli") as tmpdir:
            log.info("Created temporary directory %s", tmpdir)
            yield SecurePath(tmpdir)
            log.info("Removing temporary directory %s", tmpdir)

    def _assert_exists_and_is_file(self) -> None:
        self.assert_exists()
        self.assert_is_file()

    def assert_exists(self) -> None:
        if not self.exists():
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), self._path.resolve())

    def assert_is_file(self) -> None:
        if not self._path.is_file():
            _raise_is_a_directory_error(self._path.resolve())

    def assert_is_directory(self) -> None:
        if not self._path.is_dir():
            _raise_not_a_directory_error(self._path.resolve())

    def _assert_file_size_limit(self, size_limit_in_mb):
        if size_limit_in_mb != UNLIMITED and self._path.stat().st_size > size_limit_in_mb * 1024 * 1024:
            raise FileTooLargeError(self._path.resolve(), size_limit_in_mb)


def _raise_file_exists_error(path: Path):
    raise FileExistsError(errno.EEXIST, os.strerror(errno.EEXIST), path)


def _raise_is_a_directory_error(path: Path):
    raise IsADirectoryError(errno.EISDIR, os.strerror(errno.EISDIR), path)


def _raise_not_a_directory_error(path: Path):
    raise NotADirectoryError(errno.ENOTDIR, os.strerror(errno.ENOTDIR), path)


def _avoid_closing_the_connection_if_it_was_shared(
    using_session_token: bool, using_master_token: bool, connection_parameters: Dict
):
    if using_session_token and using_master_token:
        connection_parameters["server_session_keep_alive"] = True


def _raise_errors_related_to_session_token(using_session_token: bool, using_master_token: bool):
    if using_session_token and not using_master_token:
        raise ClickException("When using a session token, you must provide the corresponding master token")
    if using_master_token and not using_session_token:
        raise ClickException("When using a master token, you must provide the corresponding session token")


def _update_connection_application_name(connection_parameters: Dict):
    connection_application_params = {
        "application_name": "titan-core",
    }
    connection_parameters.update(connection_application_params)


def _update_connection_details_with_private_key(connection_parameters: Dict):
    if "private_key_path" in connection_parameters:
        if connection_parameters.get("authenticator") == "SNOWFLAKE_JWT":
            private_key = _load_pem_to_der(connection_parameters["private_key_path"])
            connection_parameters["private_key"] = private_key
            del connection_parameters["private_key_path"]
        else:
            raise ClickException("Private Key authentication requires authenticator set to SNOWFLAKE_JWT")
    return connection_parameters


def _load_pem_to_der(private_key_path: str) -> bytes:
    """
    Given a private key file path (in PEM format), decode key data into DER
    format
    """
    from cryptography.hazmat.backends import default_backend
    from cryptography.hazmat.primitives.serialization import (
        Encoding,
        NoEncryption,
        PrivateFormat,
        load_pem_private_key,
    )

    with SecurePath(private_key_path).open("rb", read_file_limit_mb=DEFAULT_SIZE_LIMIT_MB) as f:
        private_key_pem = f.read()

    private_key_passphrase = os.getenv("PRIVATE_KEY_PASSPHRASE", None)
    if private_key_pem.startswith(ENCRYPTED_PKCS8_PK_HEADER) and private_key_passphrase is None:
        raise ClickException(
            "Encrypted private key, you must provide the"
            "passphrase in the environment variable PRIVATE_KEY_PASSPHRASE"
        )

    if not private_key_pem.startswith(ENCRYPTED_PKCS8_PK_HEADER) and not private_key_pem.startswith(
        UNENCRYPTED_PKCS8_PK_HEADER
    ):
        raise ClickException("Private key provided is not in PKCS#8 format. Please use correct format.")

    if private_key_pem.startswith(UNENCRYPTED_PKCS8_PK_HEADER):
        private_key_passphrase = None

    private_key = load_pem_private_key(
        private_key_pem,
        (str.encode(private_key_passphrase) if private_key_passphrase is not None else private_key_passphrase),
        default_backend(),
    )

    return private_key.private_bytes(
        encoding=Encoding.DER,
        format=PrivateFormat.PKCS8,
        encryption_algorithm=NoEncryption(),
    )


def get_env_vars() -> dict:
    connection_params = {}
    for var in [
        "SNOWFLAKE_ACCOUNT",
        "SNOWFLAKE_USER",
        "SNOWFLAKE_PASSWORD",
        "SNOWFLAKE_DATABASE",
        "SNOWFLAKE_SCHEMA",
        "SNOWFLAKE_ROLE",
        "SNOWFLAKE_WAREHOUSE",
        "SNOWFLAKE_MFA_PASSCODE",
    ]:
        value = os.getenv(var, None)
        if value:
            connection_params[var[10:].lower()] = value
    return connection_params
